import json
import logging
import os
import shutil

import boto3
from botocore.exceptions import ClientError

from bentoml.exceptions import (
    YataiDeploymentException,
    AWSServiceError,
    InvalidArgument,
    BentoMLException,
)
from bentoml.saved_bundle import loader
from bentoml.utils.tempdir import TempDirectory
from bentoml.yatai.deployment.operator import DeploymentOperatorBase
from bentoml.yatai.deployment.utils import (
    raise_if_api_names_not_found_in_bento_service_metadata,
)
from bentoml.yatai.deployment.docker_utils import (
    ensure_docker_available_or_raise,
    generate_docker_image_tag,
    push_docker_image_to_repository,
    build_docker_image,
)
from bentoml.yatai.deployment.aws_utils import (
    generate_aws_compatible_string,
    get_default_aws_region,
    get_ecr_login_info,
    create_ecr_repository_if_not_exists,
    generate_bentoml_exception_from_aws_client_error,
)
from bentoml.yatai.proto.deployment_pb2 import (
    DeploymentState,
    ApplyDeploymentResponse,
    DeleteDeploymentResponse,
    DescribeDeploymentResponse,
)
from bentoml.yatai.proto.repository_pb2 import GetBentoRequest, BentoUri
from bentoml.yatai.status import Status

logger = logging.getLogger(__name__)


BENTO_SERVICE_SAGEMAKER_DOCKERFILE = """\
FROM {docker_base_image}

# the env var $PORT is required by heroku container runtime
ENV PORT 8080
EXPOSE $PORT

RUN apt-get update --fix-missing && \
    apt-get install -y nginx && \
    apt-get clean

# gevent required by AWS Sagemaker
RUN pip install gevent>=20.9.0

# copy over model files
COPY . /bento
WORKDIR /bento

RUN if [ -f /bento/bentoml-init.sh ]; then bash -c /bento/bentoml-init.sh; fi

ENV PATH="/bento:$PATH"
"""  # noqa: E501


def get_arn_role_from_current_aws_user():
    sts_client = boto3.client("sts")
    identity = sts_client.get_caller_identity()
    sts_arn = identity["Arn"]
    sts_arn_list = sts_arn.split(":")
    type_role = sts_arn_list[-1].split("/")
    iam_client = boto3.client("iam")
    if type_role[0] in ("user", "root"):
        role_list = iam_client.list_roles()
        arn = None
        for role in role_list["Roles"]:
            policy_document = role["AssumeRolePolicyDocument"]
            statement = policy_document["Statement"][0]
            if (
                "Service" in statement["Principal"]
                and statement["Effect"] == "Allow"
                and "sagemaker.amazonaws.com" in statement["Principal"]["Service"]
            ):
                arn = role["Arn"]
        if arn is None:
            raise YataiDeploymentException(
                "Can't find proper Arn role for Sagemaker, please create one and try "
                "again"
            )
        return arn
    elif type_role[0] in ["role", "assumed-role"]:
        role_response = iam_client.get_role(RoleName=type_role[1])
        return role_response["Role"]["Arn"]

    raise YataiDeploymentException(
        "Not supported role type {}; sts arn is {}".format(type_role[0], sts_arn)
    )


def create_and_push_docker_image_to_ecr(
    region, bento_name, bento_version, snapshot_path
):
    """Create BentoService sagemaker image and push to AWS ECR

    Example: https://github.com/awslabs/amazon-sagemaker-examples/blob/\
        master/advanced_functionality/scikit_bring_your_own/container/build_and_push.sh
    1. get aws account info and login ecr
    2. create ecr repository, if not exist
    3. build tag and push docker image

    Args:
        region(String)
        bento_name(String)
        bento_version(String)
        snapshot_path(Path)

    Returns:
        str: AWS ECR Tag
    """
    repository_id = create_ecr_repository_if_not_exists(
        region, f'{bento_name}-sagemaker'.lower()
    )
    logger.debug("Getting docker login info from AWS")
    registry_url, username, password = get_ecr_login_info(
        region, repository_id=repository_id
    )

    ecr_tag = generate_docker_image_tag(
        f'{bento_name}-sagemaker', bento_version, registry_url
    )

    logger.debug("Building docker image: %s", ecr_tag)
    build_docker_image(
        context_path=snapshot_path, dockerfile='Dockerfile-sagemaker', image_tag=ecr_tag
    )

    logger.debug("Pushing image to AWS ECR at %s", ecr_tag)
    push_docker_image_to_repository(
        repository=ecr_tag, username=username, password=password
    )
    logger.debug("Finished pushing image: %s", ecr_tag)
    return ecr_tag


# Sagemaker response status: 'OutOfService'|'Creating'|'Updating'|
#                            'SystemUpdating'|'RollingBack'|'InService'|
#                            'Deleting'|'Failed'
ENDPOINT_STATUS_TO_STATE = {
    "InService": DeploymentState.RUNNING,
    "Deleting": DeploymentState.INACTIVATED,
    "Creating": DeploymentState.PENDING,
    "Updating": DeploymentState.PENDING,
    "RollingBack": DeploymentState.PENDING,
    "SystemUpdating": DeploymentState.PENDING,
    "OutOfService": DeploymentState.INACTIVATED,
    "Failed": DeploymentState.ERROR,
}


def _aws_client_error_to_bentoml_exception(e, message_prefix=None):
    """parse botocore.exceptions.ClientError into Bento StatusProto

    We handle two most common errors when deploying to Sagemaker.
        1. Authentication issue/invalid access(InvalidSignatureException)
        2. resources not found (ValidationException)
    It will return correlated StatusProto(NOT_FOUND, UNAUTHENTICATED)

    Args:
        e: ClientError from botocore.exceptions
    Returns:
        StatusProto
    """
    error_response = e.response.get("Error", {})
    error_code = error_response.get("Code", "Unknown")
    error_message = error_response.get("Message", "Unknown")
    error_log_message = (
        f"AWS ClientError - operation: {e.operation_name}, "
        f"code: {error_code}, message: {error_message}"
    )
    if message_prefix:
        error_log_message = f"{message_prefix}; {error_log_message}"
    logger.error(error_log_message)
    return AWSServiceError(error_log_message)


def _get_sagemaker_resource_names(deployment_pb):
    sagemaker_model_name = generate_aws_compatible_string(
        (deployment_pb.namespace, 10),
        (deployment_pb.name, 12),
        (deployment_pb.spec.bento_name, 20),
        (deployment_pb.spec.bento_version, 18),
    )
    sagemaker_endpoint_config_name = generate_aws_compatible_string(
        (deployment_pb.namespace, 10),
        (deployment_pb.name, 12),
        (deployment_pb.spec.bento_name, 20),
        (deployment_pb.spec.bento_version, 18),
    )
    sagemaker_endpoint_name = generate_aws_compatible_string(
        deployment_pb.namespace, deployment_pb.name
    )
    return sagemaker_model_name, sagemaker_endpoint_config_name, sagemaker_endpoint_name


def _delete_sagemaker_model_if_exist(sagemaker_client, sagemaker_model_name):
    try:
        delete_model_response = sagemaker_client.delete_model(
            ModelName=sagemaker_model_name
        )
        logger.debug("AWS delete model response: %s", delete_model_response)
    except ClientError as e:
        error_response = e.response.get("Error", {})
        error_code = error_response.get("Code", "Unknown")
        error_message = error_response.get("Message", "Unknown")
        if (
            error_code == "ValidationException"
            and "Could not find model" in error_message
        ):
            # sagemaker model does not exist
            return

        raise generate_bentoml_exception_from_aws_client_error(
            e, f"Failed to cleanup sagemaker model '{sagemaker_model_name}'"
        )

    return


def _delete_sagemaker_endpoint_config_if_exist(
    sagemaker_client, sagemaker_endpoint_config_name
):
    try:
        delete_endpoint_config_response = sagemaker_client.delete_endpoint_config(
            EndpointConfigName=sagemaker_endpoint_config_name
        )
        logger.debug(
            "AWS delete endpoint config response: %s", delete_endpoint_config_response
        )
    except ClientError as e:
        error_response = e.response.get("Error", {})
        error_code = error_response.get("Code", "Unknown")
        error_message = error_response.get("Message", "Unknown")
        if (
            error_code == "ValidationException"
            and "Could not find endpoint configuration" in error_message
        ):
            # endpoint config does not exist
            return

        raise generate_bentoml_exception_from_aws_client_error(
            e,
            f"Failed to cleanup sagemaker endpoint config "
            f"'{sagemaker_endpoint_config_name}' after creation failed",
        )
    return


def _delete_sagemaker_endpoint_if_exist(sagemaker_client, sagemaker_endpoint_name):
    try:
        delete_endpoint_response = sagemaker_client.delete_endpoint(
            EndpointName=sagemaker_endpoint_name
        )
        logger.debug("AWS delete endpoint response: %s", delete_endpoint_response)
    except ClientError as e:
        error_response = e.response.get("Error", {})
        error_code = error_response.get("Code", "Unknown")
        error_message = error_response.get("Message", "Unknown")
        if (
            error_code == "ValidationException"
            and "Could not find endpoint" in error_message
        ):
            # sagemaker endpoint does not exist
            return

        raise generate_bentoml_exception_from_aws_client_error(
            e, f"Failed to delete sagemaker endpoint '{sagemaker_endpoint_name}'"
        )


def delete_sagemaker_deployment_resources_if_exist(deployment_pb):
    sagemaker_config = deployment_pb.spec.sagemaker_operator_config
    sagemaker_client = boto3.client("sagemaker", sagemaker_config.region)

    (
        sagemaker_model_name,
        sagemaker_endpoint_config_name,
        sagemaker_endpoint_name,
    ) = _get_sagemaker_resource_names(deployment_pb)

    _delete_sagemaker_model_if_exist(sagemaker_client, sagemaker_model_name)
    _delete_sagemaker_endpoint_config_if_exist(
        sagemaker_client, sagemaker_endpoint_config_name
    )
    _delete_sagemaker_endpoint_if_exist(sagemaker_client, sagemaker_endpoint_name)


def _init_sagemaker_project(sagemaker_project_dir, bento_path, docker_base_image):
    shutil.copytree(bento_path, sagemaker_project_dir)

    with open(os.path.join(sagemaker_project_dir, "Dockerfile-sagemaker"), "w") as f:
        f.write(
            BENTO_SERVICE_SAGEMAKER_DOCKERFILE.format(
                docker_base_image=docker_base_image
            )
        )

    nginx_conf_path = os.path.join(os.path.dirname(__file__), "nginx.conf")
    shutil.copy(nginx_conf_path, os.path.join(sagemaker_project_dir, "nginx.conf"))

    wsgi_py_path = os.path.join(os.path.dirname(__file__), "wsgi.py")
    shutil.copy(wsgi_py_path, os.path.join(sagemaker_project_dir, "wsgi.py"))

    serve_file_path = os.path.join(os.path.dirname(__file__), "serve")
    shutil.copy(serve_file_path, os.path.join(sagemaker_project_dir, "serve"))

    # permission 755 is required for entry script 'serve'
    os.chmod(os.path.join(sagemaker_project_dir, "serve"), 0o755)
    return sagemaker_project_dir


def _create_sagemaker_model(
    sagemaker_client, sagemaker_model_name, ecr_image_path, spec
):
    execution_role_arn = get_arn_role_from_current_aws_user()

    sagemaker_model_info = {
        "ModelName": sagemaker_model_name,
        "PrimaryContainer": {
            "ContainerHostname": sagemaker_model_name,
            "Image": ecr_image_path,
            "Environment": {
                "API_NAME": spec.api_name,
                "BENTOML_GUNICORN_TIMEOUT": str(spec.timeout),
            },
        },
        "ExecutionRoleArn": execution_role_arn,
    }

    # Will set envvar, if user defined gunicorn workers per instance.  EnvVar needs
    # to be string instead of the int.
    if spec.num_of_gunicorn_workers_per_instance:
        sagemaker_model_info["PrimaryContainer"]["Environment"][
            "BENTOML_GUNICORN_NUM_OF_WORKERS"
        ] = str(spec.num_of_gunicorn_workers_per_instance)

    try:
        create_model_response = sagemaker_client.create_model(**sagemaker_model_info)
    except ClientError as e:
        raise generate_bentoml_exception_from_aws_client_error(
            e, "Failed to create sagemaker model"
        )
    logger.debug("AWS create model response: %s", create_model_response)


def _create_sagemaker_endpoint_config(
    sagemaker_client, sagemaker_model_name, endpoint_config_name, sagemaker_config
):
    create_endpoint_config_arguments = {
        "EndpointConfigName": endpoint_config_name,
        "ProductionVariants": [
            {
                "VariantName": sagemaker_model_name,
                "ModelName": sagemaker_model_name,
                "InitialInstanceCount": sagemaker_config.instance_count,
                "InstanceType": sagemaker_config.instance_type,
            }
        ],
    }

    if sagemaker_config.data_capture_s3_prefix:
        logger.debug(
            "data_capture_s3_prefix %s found, creating data capture config",
            sagemaker_config.data_capture_s3_prefix,
        )

        if not sagemaker_config.data_capture_sample_percent:
            logger.debug("data_capture_sample_percent not found, using default 100%")
            sagemaker_config.data_capture_sample_percent = 100

        create_endpoint_config_arguments["DataCaptureConfig"] = {
            "EnableCapture": True,
            "DestinationS3Uri": sagemaker_config.data_capture_s3_prefix,
            "InitialSamplingPercentage": sagemaker_config.data_capture_sample_percent,
            "CaptureOptions": [{"CaptureMode": "Input"}, {"CaptureMode": "Output"}],
        }

    logger.debug("Creating Sagemaker endpoint %s configuration", endpoint_config_name)
    try:
        create_config_response = sagemaker_client.create_endpoint_config(
            **create_endpoint_config_arguments
        )
    except ClientError as e:
        raise generate_bentoml_exception_from_aws_client_error(
            e, "Failed to create sagemaker endpoint config"
        )
    logger.debug("AWS create endpoint config response: %s", create_config_response)


def _create_sagemaker_endpoint(sagemaker_client, endpoint_name, endpoint_config_name):
    try:
        logger.debug("Creating sagemaker endpoint %s", endpoint_name)
        create_endpoint_response = sagemaker_client.create_endpoint(
            EndpointName=endpoint_name, EndpointConfigName=endpoint_config_name
        )
        logger.debug("AWS create endpoint response: %s", create_endpoint_response)
    except ClientError as e:
        raise generate_bentoml_exception_from_aws_client_error(
            e, "Failed to create sagemaker endpoint"
        )


def _update_sagemaker_endpoint(sagemaker_client, endpoint_name, endpoint_config_name):
    try:
        logger.debug("Updating sagemaker endpoint %s", endpoint_name)
        update_endpoint_response = sagemaker_client.update_endpoint(
            EndpointName=endpoint_name, EndpointConfigName=endpoint_config_name
        )
        logger.debug("AWS update endpoint response: %s", str(update_endpoint_response))
    except ClientError as e:
        raise generate_bentoml_exception_from_aws_client_error(
            e, "Failed to update sagemaker endpoint"
        )


class SageMakerDeploymentOperator(DeploymentOperatorBase):
    def __init__(self, yatai_service):
        super(SageMakerDeploymentOperator, self).__init__(yatai_service)
        ensure_docker_available_or_raise()

    def add(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            sagemaker_config = deployment_spec.sagemaker_operator_config
            sagemaker_config.region = (
                sagemaker_config.region or get_default_aws_region()
            )

            if sagemaker_config is None:
                raise YataiDeploymentException("Sagemaker configuration is missing.")

            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )
            if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
                raise BentoMLException(
                    "BentoML currently not support {} repository".format(
                        BentoUri.StorageType.Name(bento_pb.bento.uri.type)
                    )
                )
            return self._add(deployment_pb, bento_pb, bento_pb.bento.uri.uri)

        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = (
                f"Error creating SageMaker deployment: {str(error)}"
            )
            return ApplyDeploymentResponse(
                status=error.status_proto, deployment=deployment_pb
            )

    def _add(self, deployment_pb, bento_pb, bento_path):
        if loader._is_remote_path(bento_path):
            with loader._resolve_remote_bundle_path(bento_path) as local_path:
                return self._add(deployment_pb, bento_pb, local_path)

        deployment_spec = deployment_pb.spec
        sagemaker_config = deployment_spec.sagemaker_operator_config

        raise_if_api_names_not_found_in_bento_service_metadata(
            bento_pb.bento.bento_service_metadata, [sagemaker_config.api_name]
        )

        sagemaker_client = boto3.client("sagemaker", sagemaker_config.region)

        with TempDirectory() as temp_dir:
            sagemaker_project_dir = os.path.join(temp_dir, deployment_spec.bento_name)
            _init_sagemaker_project(
                sagemaker_project_dir,
                bento_path,
                bento_pb.bento.bento_service_metadata.env.docker_base_image,
            )
            ecr_image_path = create_and_push_docker_image_to_ecr(
                sagemaker_config.region,
                deployment_spec.bento_name,
                deployment_spec.bento_version,
                sagemaker_project_dir,
            )

        try:
            (
                sagemaker_model_name,
                sagemaker_endpoint_config_name,
                sagemaker_endpoint_name,
            ) = _get_sagemaker_resource_names(deployment_pb)

            _create_sagemaker_model(
                sagemaker_client, sagemaker_model_name, ecr_image_path, sagemaker_config
            )
            _create_sagemaker_endpoint_config(
                sagemaker_client,
                sagemaker_model_name,
                sagemaker_endpoint_config_name,
                sagemaker_config,
            )
            _create_sagemaker_endpoint(
                sagemaker_client,
                sagemaker_endpoint_name,
                sagemaker_endpoint_config_name,
            )
        except AWSServiceError as e:
            delete_sagemaker_deployment_resources_if_exist(deployment_pb)
            raise e

        return ApplyDeploymentResponse(status=Status.OK(), deployment=deployment_pb)

    def update(self, deployment_pb, previous_deployment):
        try:
            deployment_spec = deployment_pb.spec
            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )
            if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
                raise BentoMLException(
                    "BentoML currently not support {} repository".format(
                        BentoUri.StorageType.Name(bento_pb.bento.uri.type)
                    )
                )
            return self._update(
                deployment_pb, previous_deployment, bento_pb, bento_pb.bento.uri.uri
            )
        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = (
                f"Error updating SageMaker deployment: {str(error)}"
            )
            return ApplyDeploymentResponse(
                status=error.status_proto, deployment=deployment_pb
            )

    def _update(self, deployment_pb, current_deployment, bento_pb, bento_path):
        if loader._is_remote_path(bento_path):
            with loader._resolve_remote_bundle_path(bento_path) as local_path:
                return self._update(
                    deployment_pb, current_deployment, bento_pb, local_path
                )
        updated_deployment_spec = deployment_pb.spec
        updated_sagemaker_config = updated_deployment_spec.sagemaker_operator_config
        sagemaker_client = boto3.client(
            "sagemaker", updated_sagemaker_config.region or get_default_aws_region()
        )

        try:
            raise_if_api_names_not_found_in_bento_service_metadata(
                bento_pb.bento.bento_service_metadata,
                [updated_sagemaker_config.api_name],
            )
            describe_latest_deployment_state = self.describe(deployment_pb)
            current_deployment_spec = current_deployment.spec
            current_sagemaker_config = current_deployment_spec.sagemaker_operator_config
            latest_deployment_state = json.loads(
                describe_latest_deployment_state.state.info_json
            )

            current_ecr_image_tag = latest_deployment_state["ProductionVariants"][0][
                "DeployedImages"
            ][0]["SpecifiedImage"]
            if (
                updated_deployment_spec.bento_name != current_deployment_spec.bento_name
                or updated_deployment_spec.bento_version
                != current_deployment_spec.bento_version
            ):
                logger.debug(
                    "BentoService tag is different from current deployment, "
                    "creating new docker image and push to ECR"
                )
                with TempDirectory() as temp_dir:
                    sagemaker_project_dir = os.path.join(
                        temp_dir, updated_deployment_spec.bento_name
                    )
                    _init_sagemaker_project(
                        sagemaker_project_dir,
                        bento_path,
                        bento_pb.bento.bento_service_metadata.env.docker_base_image,
                    )
                    ecr_image_path = create_and_push_docker_image_to_ecr(
                        updated_sagemaker_config.region,
                        updated_deployment_spec.bento_name,
                        updated_deployment_spec.bento_version,
                        sagemaker_project_dir,
                    )
            else:
                logger.debug("Using existing ECR image for Sagemaker model")
                ecr_image_path = current_ecr_image_tag

            (
                updated_sagemaker_model_name,
                updated_sagemaker_endpoint_config_name,
                sagemaker_endpoint_name,
            ) = _get_sagemaker_resource_names(deployment_pb)
            (
                current_sagemaker_model_name,
                current_sagemaker_endpoint_config_name,
                _,
            ) = _get_sagemaker_resource_names(current_deployment)

            if (
                updated_sagemaker_config.api_name != current_sagemaker_config.api_name
                or updated_sagemaker_config.num_of_gunicorn_workers_per_instance
                != current_sagemaker_config.num_of_gunicorn_workers_per_instance
                or ecr_image_path != current_ecr_image_tag
            ):
                logger.debug(
                    "Sagemaker model requires update. Delete current sagemaker model %s"
                    "and creating new model %s",
                    current_sagemaker_model_name,
                    updated_sagemaker_model_name,
                )
                _delete_sagemaker_model_if_exist(
                    sagemaker_client, current_sagemaker_model_name
                )
                _create_sagemaker_model(
                    sagemaker_client,
                    updated_sagemaker_model_name,
                    ecr_image_path,
                    updated_sagemaker_config,
                )
            # When bento service tag is not changed, we need to delete the current
            # endpoint configuration in order to create new one to avoid name collation
            if (
                current_sagemaker_endpoint_config_name
                == updated_sagemaker_endpoint_config_name
            ):
                logger.debug(
                    "Current sagemaker config name %s is same as updated one, "
                    "delete it before create new endpoint config",
                    current_sagemaker_endpoint_config_name,
                )
                _delete_sagemaker_endpoint_config_if_exist(
                    sagemaker_client, current_sagemaker_endpoint_config_name
                )
            logger.debug(
                "Create new endpoint configuration %s",
                updated_sagemaker_endpoint_config_name,
            )
            _create_sagemaker_endpoint_config(
                sagemaker_client,
                updated_sagemaker_model_name,
                updated_sagemaker_endpoint_config_name,
                updated_sagemaker_config,
            )
            logger.debug(
                "Updating endpoint to new endpoint configuration %s",
                updated_sagemaker_endpoint_config_name,
            )
            _update_sagemaker_endpoint(
                sagemaker_client,
                sagemaker_endpoint_name,
                updated_sagemaker_endpoint_config_name,
            )
            if not (
                current_sagemaker_endpoint_config_name
                == updated_sagemaker_endpoint_config_name
            ):
                logger.debug(
                    'Delete old sagemaker endpoint config %s',
                    current_sagemaker_endpoint_config_name,
                )
                _delete_sagemaker_endpoint_config_if_exist(
                    sagemaker_client, current_sagemaker_endpoint_config_name
                )
        except AWSServiceError as e:
            delete_sagemaker_deployment_resources_if_exist(deployment_pb)
            raise e

        return ApplyDeploymentResponse(status=Status.OK(), deployment=deployment_pb)

    def delete(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            sagemaker_config = deployment_spec.sagemaker_operator_config
            sagemaker_config.region = (
                sagemaker_config.region or get_default_aws_region()
            )
            if not sagemaker_config.region:
                raise InvalidArgument("AWS region is missing")

            delete_sagemaker_deployment_resources_if_exist(deployment_pb)

            return DeleteDeploymentResponse(status=Status.OK())
        except BentoMLException as error:
            return DeleteDeploymentResponse(status=error.status_proto)

    def describe(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            sagemaker_config = deployment_spec.sagemaker_operator_config
            sagemaker_config.region = (
                sagemaker_config.region or get_default_aws_region()
            )
            if not sagemaker_config.region:
                raise InvalidArgument("AWS region is missing")
            sagemaker_client = boto3.client("sagemaker", sagemaker_config.region)
            _, _, sagemaker_endpoint_name = _get_sagemaker_resource_names(deployment_pb)

            try:
                endpoint_status_response = sagemaker_client.describe_endpoint(
                    EndpointName=sagemaker_endpoint_name
                )
            except ClientError as e:
                raise _aws_client_error_to_bentoml_exception(
                    e,
                    f"Failed to fetch current status of sagemaker endpoint "
                    f"'{sagemaker_endpoint_name}'",
                )

            logger.debug("AWS describe endpoint response: %s", endpoint_status_response)
            endpoint_status = endpoint_status_response["EndpointStatus"]

            service_state = ENDPOINT_STATUS_TO_STATE[endpoint_status]

            deployment_state = DeploymentState(
                state=service_state,
                info_json=json.dumps(endpoint_status_response, default=str),
            )
            deployment_state.timestamp.GetCurrentTime()

            return DescribeDeploymentResponse(
                state=deployment_state, status=Status.OK()
            )
        except BentoMLException as error:
            return DescribeDeploymentResponse(status=error.status_proto)



import gzip
import json
import os

from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer
from bentoml.types import HTTPRequest, InferenceTask

TF_B64_KEY = "b64"


class B64JsonEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    def default(self, o):  # pylint: disable=method-hidden
        import base64

        if isinstance(o, bytes):
            try:
                return o.decode('utf-8')
            except UnicodeDecodeError:
                return {TF_B64_KEY: base64.b64encode(o).decode("utf-8")}

        try:
            return super(B64JsonEncoder, self).default(o)
        except (TypeError, OverflowError):
            return {"unknown_obj": str(o)}


class NumpyJsonEncoder(B64JsonEncoder):
    """ Special json encoder for numpy types """

    def default(self, o):  # pylint: disable=method-hidden
        import numpy as np

        if isinstance(o, np.generic):
            return o.item()

        if isinstance(o, np.ndarray):
            return o.tolist()

        return super(NumpyJsonEncoder, self).default(o)


class TfTensorJsonEncoder(NumpyJsonEncoder):
    """ Special json encoder for numpy types """

    def default(self, o):  # pylint: disable=method-hidden

        import tensorflow as tf

        # Tensor -> ndarray or object
        if isinstance(o, tf.Tensor):
            if tf.__version__.startswith("1."):
                with tf.compat.v1.Session():
                    return o.numpy()
            else:
                return o.numpy()
        return super(TfTensorJsonEncoder, self).default(o)


def concat_list(lst, batch_flags=None):
    """
    >>> lst = [
        [1],
        [1, 2],
        [1, 2, 3],
        None,
        ]
    >>> concat_list(lst)
    [1, 1, 2, 1, 2, 3], [slice(0, 1), slice(1, 3), slice(3, 6), None]
    """
    slices = [slice(0)] * len(lst)
    datas = []
    row_flag = 0
    for i, r in enumerate(lst):
        if r is None:
            slices[i] = None
            continue
        j = -1
        if batch_flags is None or batch_flags[i]:
            for j, d in enumerate(r):
                datas.append(d)
            slices[i] = slice(row_flag, row_flag + j + 1)
        else:
            datas.append(r)
            slices[i] = row_flag
            j += 1
        row_flag += j + 1
    return datas, slices


def check_file_extension(file_name, accept_ext_list):
    """
    Return False if file's extension is not in the accept_ext_list
    """
    if not file_name:
        return False
    _, extension = os.path.splitext(file_name)
    return extension.lower() in (accept_ext_list or [])


@inject
def get_default_accept_image_formats(
    default_extensions: list = Provide[
        BentoMLContainer.config.adapters.image_input.default_extensions
    ],
):
    """With default bentoML config, this returns:
        ['.jpg', '.png', '.jpeg', '.tiff', '.webp', '.bmp']
    """
    return default_extensions


def decompress_gzip_request(method):
    def _method(self, req: HTTPRequest) -> InferenceTask:
        if req.headers.content_encoding in {"gzip", "x-gzip"}:
            # https://tools.ietf.org/html/rfc7230#section-4.2.3
            try:
                req.body = gzip.decompress(req.body)

            except OSError:
                return InferenceTask().discard(
                    http_status=400, err_msg="Gzip decompression error"
                )
            req.headers.pop("content-encoding")
            return method(self, req)
        return method(self, req)

    return _method



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from bentoml.exceptions import BentoMLException
from bentoml.yatai.proto.label_selectors_pb2 import LabelSelectors


label_expression_operators = {
    i[0]: i[1] for i in LabelSelectors.LabelSelectorExpression.OPERATOR_TYPE.items()
}


def _extract_expressions(query):
    # Using regex to split query base on ",".  negative lookahead (?!...) / don't match
    # anything inside the ().
    # e.g.
    # "key, key1 in (value1, value2),key3"
    # -> ['key', ' key1 in (value1, value2)', 'key3']
    return re.split(r',(?![^()]*\))', query)


def value_string_to_list(value_string):
    if not value_string.startswith('(') or not value_string.endswith(')'):
        raise BentoMLException(
            f"Query values {value_string} need to be inside (). "
            f"e.g. (value1, value2, ..)"
        )
    if len(value_string) == 2:
        raise BentoMLException("Query values can't be empty")
    return [value.strip() for value in value_string[1:-1].split(',')]


def _extract_expression_elements(expression):
    # Using regex to split query base on " ".  negative lookahead (?!...) / don't match
    # anything inside the ().
    # e.g.
    # "key1 in (value1, value2,value4)"
    # -> ['key1', ' in', '(value1, value2,,value3)']
    return re.split(r' (?![^()]*\))', expression)


def generate_gprc_labels_selector(label_selectors, label_query):
    """Parse label query string and update to the label selector request"""
    if not label_query:
        return
    expressions = _extract_expressions(label_query)
    for expression in expressions:
        expression = expression.strip()
        elements = _extract_expression_elements(expression.strip())
        if len(elements) == 1:
            # Possible queries: key=value, key!=value, key
            query = elements[0].strip()
            if query.lower() in [i.lower() for i in label_expression_operators.keys()]:
                raise BentoMLException(
                    f"Label query operator {query} can't be the only element"
                )
            if '!=' in query:
                if query.count('!=') > 1:
                    raise BentoMLException(f"Too many '!=' operator in query {query}")
                key, value = query.split('!=')
                if not value:
                    raise BentoMLException(f"Label {query} can't have empty value")
                label_selectors.match_expressions.append(
                    LabelSelectors.LabelSelectorExpression(
                        key=key,
                        operator=label_expression_operators['NotIn'],
                        values=[value],
                    )
                )
            elif '=' in query:
                if query.count('=') > 1:
                    raise BentoMLException(f"Too many '=' operator in query {query}")
                key, value = query.split('=')
                if not value:
                    raise BentoMLException(f"Label {query} can't have empty value")
                label_selectors.match_labels[key] = value
            else:
                label_selectors.match_expressions.append(
                    LabelSelectors.LabelSelectorExpression(
                        key=query, operator=label_expression_operators['Exists']
                    )
                )
        elif len(elements) == 2:
            # possible queries: key Exists/exists, key DoesNotExist/doesnotexist
            key = elements[0].strip()
            operator = elements[1].strip()
            assert operator.lower() in [
                i.lower() for i in ['Exists', 'DoesNotExist']
            ], f'Operator "{operator}" is invalid'
            operator = 'Exists' if operator.capitalize() == 'Exists' else 'DoesNotExist'
            label_selectors.match_expressions.append(
                LabelSelectors.LabelSelectorExpression(
                    key=key, operator=label_expression_operators[operator],
                )
            )
        elif len(elements) == 3:
            # possible queries: key In/in (values), key NotIn/notin (values)
            key = elements[0].strip()
            operator = elements[1].strip()
            assert operator.lower() in [
                i.lower() for i in ['In', 'NotIn']
            ], f'Operator "{operator}" is invalid'
            operator = 'In' if operator.capitalize() == 'In' else 'NotIn'
            values = value_string_to_list(elements[2].strip())
            label_selectors.match_expressions.append(
                LabelSelectors.LabelSelectorExpression(
                    key=key,
                    operator=label_expression_operators[operator],
                    values=values,
                )
            )
        else:
            raise BentoMLException(f'Too many elements in the label query {expression}')
    return



import logging

logger = logging.getLogger(__name__)
logger.warning(
    """\
Importing from "bentoml.artifact.*" has been deprecated. Instead, use\
`bentoml.frameworks.*` and `bentoml.service.*`. e.g.:, \
`from bentoml.frameworks.sklearn import SklearnModelArtifact`, \
`from bentoml.service.artifacts import BentoServiceArtifact`, \
`from bentoml.service.artifacts.common import PickleArtifact`"""
)

from bentoml.service.artifacts import (
    BentoServiceArtifact,
    ArtifactCollection,
)

from bentoml.service.artifacts.common import TextFileArtifact
from bentoml.service.artifacts.common import JSONArtifact
from bentoml.service.artifacts.common import PickleArtifact

from bentoml.frameworks.coreml import CoreMLModelArtifact
from bentoml.frameworks.detectron import DetectronModelArtifact
from bentoml.frameworks.evalml import EvalMLModelArtifact  # noqa: E402
from bentoml.frameworks.fastai import (  # noqa: E402
    FastaiModelArtifact as Fastai2ModelArtifact,
)  # noqa: E402
from bentoml.frameworks.fastai import (  # noqa: E402
    Fastai1ModelArtifact as FastaiModelArtifact,
)  # noqa: E402
from bentoml.frameworks.fasttext import FasttextModelArtifact
from bentoml.frameworks.h2o import H2oModelArtifact
from bentoml.frameworks.keras import KerasModelArtifact
from bentoml.frameworks.lightgbm import LightGBMModelArtifact
from bentoml.frameworks.onnx import OnnxModelArtifact
from bentoml.frameworks.onnxmlir import OnnxMlirModelArtifact
from bentoml.frameworks.pytorch import (  # noqa: E402
    PytorchModelArtifact,
    PytorchLightningModelArtifact,
)  # noqa: E402
from bentoml.frameworks.sklearn import SklearnModelArtifact
from bentoml.frameworks.spacy import SpacyModelArtifact
from bentoml.frameworks.tensorflow import TensorflowSavedModelArtifact
from bentoml.frameworks.xgboost import XgboostModelArtifact
from bentoml.frameworks.paddle import PaddlePaddleModelArtifact  # noqa: E402
from bentoml.frameworks.easyocr import EasyOCRArtifact  # noqa: E402

__all__ = [
    "ArtifactCollection",
    "BentoServiceArtifact",
    "CoreMLModelArtifact",
    "DetectronModelArtifact",
    "EvalMLModelArtifact",
    "Fastai2ModelArtifact",
    "FastaiModelArtifact",
    "FasttextModelArtifact",
    "H2oModelArtifact",
    "JSONArtifact",
    "KerasModelArtifact",
    "LightGBMModelArtifact",
    "OnnxModelArtifact",
    "OnnxMlirModelArtifact",
    "PickleArtifact",
    "PytorchModelArtifact",
    "SklearnModelArtifact",
    "SpacyModelArtifact",
    "TensorflowSavedModelArtifact",
    "TextFileArtifact",
    "XgboostModelArtifact",
    "PytorchLightningModelArtifact",
    "PaddlePaddleModelArtifact",
    "EasyOCRArtifact",
]



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
import threading
import itertools
import time
import logging
from datetime import datetime

import humanfriendly
from tabulate import tabulate

from bentoml.cli.click_utils import _echo
from bentoml.utils import pb_to_yaml
from bentoml.exceptions import BentoMLException

logger = logging.getLogger(__name__)


class Spinner:
    def __init__(self, message, delay=0.1):
        self.spinner = itertools.cycle(['-', '/', '|', '\\'])
        self.delay = delay
        self.busy = False
        self._screen_lock = None
        self.thread = None
        self.spinner_visible = False
        sys.stdout.write(message)

    def write_next(self):
        with self._screen_lock:
            if not self.spinner_visible:
                sys.stdout.write(next(self.spinner))
                self.spinner_visible = True
                sys.stdout.flush()

    def remove_spinner(self, cleanup=False):
        with self._screen_lock:
            if self.spinner_visible:
                sys.stdout.write('\b')
                self.spinner_visible = False
                if cleanup:
                    sys.stdout.write(' ')  # overwrite spinner with blank
                    sys.stdout.write('\r')  # move to next line
                sys.stdout.flush()

    def spinner_task(self):
        while self.busy:
            self.write_next()
            time.sleep(self.delay)
            self.remove_spinner()

    def __enter__(self):
        if sys.stdout.isatty():
            self._screen_lock = threading.Lock()
            self.busy = True
            self.thread = threading.Thread(target=self.spinner_task)
            self.thread.start()

    def __exit__(self, exception, value, tb):
        if sys.stdout.isatty():
            self.busy = False
            self.remove_spinner(cleanup=True)
        else:
            sys.stdout.write('\r')


def parse_key_value_pairs(key_value_pairs_str):
    result = {}
    if key_value_pairs_str:
        for key_value_pair in key_value_pairs_str.split(','):
            key, value = key_value_pair.split('=')
            key = key.strip()
            value = value.strip()
            if key in result:
                logger.warning("duplicated key '%s' found string map parameter", key)
            result[key] = value
    return result


def echo_docker_api_result(docker_generator):
    layers = {}
    for line in docker_generator:
        if "stream" in line:
            cleaned = line['stream'].rstrip("\n")
            if cleaned != "":
                yield cleaned
        if "status" in line and line["status"] == "Pushing":
            progress = line["progressDetail"]
            layers[line["id"]] = progress["current"], progress["total"]
            cur, total = zip(*layers.values())
            cur, total = (
                humanfriendly.format_size(sum(cur)),
                humanfriendly.format_size(sum(total)),
            )
            yield f"Pushed {cur} / {total}"
        if "errorDetail" in line:
            error = line["errorDetail"]
            raise BentoMLException(error["message"])


def _print_deployment_info(deployment, output_type):
    if output_type == 'yaml':
        _echo(pb_to_yaml(deployment))
    else:
        from google.protobuf.json_format import MessageToDict

        deployment_info = MessageToDict(deployment)
        if deployment_info['state'] and deployment_info['state']['infoJson']:
            deployment_info['state']['infoJson'] = json.loads(
                deployment_info['state']['infoJson']
            )
        _echo(json.dumps(deployment_info, indent=2, separators=(',', ': ')))


def _format_labels_for_print(labels):
    if not labels:
        return None
    result = [f'{label_key}:{labels[label_key]}' for label_key in labels]
    return '\n'.join(result)


def _format_deployment_age_for_print(deployment_pb):
    if not deployment_pb.created_at:
        # deployments created before version 0.4.5 don't have created_at field,
        # we will not show the age for those deployments
        return None
    else:
        return human_friendly_age_from_datetime(deployment_pb.created_at.ToDatetime())


def human_friendly_age_from_datetime(dt, detailed=False, max_unit=2):
    return humanfriendly.format_timespan(datetime.utcnow() - dt, detailed, max_unit)


def _print_deployments_table(deployments, wide=False):
    from bentoml.yatai.proto.deployment_pb2 import DeploymentState, DeploymentSpec

    table = []
    if wide:
        headers = [
            'NAME',
            'NAMESPACE',
            'PLATFORM',
            'BENTO_SERVICE',
            'STATUS',
            'AGE',
            'LABELS',
        ]
    else:
        headers = [
            'NAME',
            'NAMESPACE',
            'PLATFORM',
            'BENTO_SERVICE',
            'STATUS',
            'AGE',
        ]
    for deployment in deployments:
        row = [
            deployment.name,
            deployment.namespace,
            DeploymentSpec.DeploymentOperator.Name(deployment.spec.operator)
            .lower()
            .replace('_', '-'),
            f'{deployment.spec.bento_name}:{deployment.spec.bento_version}',
            DeploymentState.State.Name(deployment.state.state)
            .lower()
            .replace('_', ' '),
            _format_deployment_age_for_print(deployment),
        ]
        if wide:
            row.append(_format_labels_for_print(deployment.labels))
        table.append(row)
    table_display = tabulate(table, headers, tablefmt='plain')
    _echo(table_display)


def _print_deployments_info(deployments, output_type):
    if output_type == 'table':
        _print_deployments_table(deployments)
    elif output_type == 'wide':
        _print_deployments_table(deployments, wide=True)
    else:
        for deployment in deployments:
            _print_deployment_info(deployment, output_type)



import json
from typing import Sequence

from bentoml.adapters.json_output import JsonOutput
from bentoml.types import InferenceError, InferenceResult, InferenceTask
from bentoml.utils.dataframe_util import PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS


def df_to_json(result, pandas_dataframe_orient="records"):
    import pandas as pd

    assert (
        pandas_dataframe_orient in PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS
    ), f"unknown pandas dataframe orient '{pandas_dataframe_orient}'"

    if isinstance(result, pd.DataFrame):
        return result.to_json(orient=pandas_dataframe_orient)

    if isinstance(result, pd.Series):
        return pd.DataFrame(result).to_json(orient=pandas_dataframe_orient)
    return json.dumps(result)


class DataframeOutput(JsonOutput):
    """
    Output adapters converts returns of user defined API function into specific output,
    such as HTTP response, command line stdout or AWS Lambda event object.

    Args:
        cors (str): DEPRECATED. Moved to the configuration file.
            The value of the Access-Control-Allow-Origin header set in the
            HTTP/AWS Lambda response object. If set to None, the header will not be set.
            Default is None.
        ensure_ascii(bool): Escape all non-ASCII characters. Default False.
        output_orient(str): The orient format of output dataframes.
            Same as pd.DataFrame.to_json(orient=orient)
    """

    BATCH_MODE_SUPPORTED = True

    def __init__(self, output_orient='records', **kwargs):
        super().__init__(**kwargs)
        self.output_orient = output_orient

        assert self.output_orient in PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS, (
            f"Invalid 'output_orient'='{self.orient}', valid options are "
            f"{PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS}"
        )

    @property
    def config(self):
        base_config = super(DataframeOutput, self).config
        return dict(base_config, output_orient=self.output_orient,)

    @property
    def pip_dependencies(self):
        """
        :return: List of PyPI package names required by this OutputAdapter
        """
        return ['pandas']

    def pack_user_func_return_value(
        self, return_result, tasks: Sequence[InferenceTask],
    ) -> Sequence[InferenceResult[str]]:
        rv = []
        i = 0
        for task in tasks:
            if task.batch is None:
                result = return_result[i : i + 1]
                i += 1
            else:
                result = return_result[i : i + task.batch]
                i += task.batch
            try:
                result = df_to_json(result, self.output_orient)
                rv.append(InferenceResult(http_status=200, data=result))
            except Exception as e:  # pylint: disable=broad-except
                rv.append(InferenceError(err_msg=str(e), http_status=500))
        return rv



import os

from bentoml.service import BentoServiceArtifact
from bentoml.utils import cloudpickle


PICKLE_FILE_EXTENSION = ".pkl"


class PickleArtifact(BentoServiceArtifact):
    """Abstraction for saving/loading python objects with pickle serialization

    Args:
        name (str): Name for the artifact
        pickle_module (module|str): The python module will be used for pickle
            and unpickle artifact, default pickle module in BentoML's fork of
            `cloudpickle`, which is identical to the Apache Spark fork
        pickle_extension (str): The extension format for pickled file.
    """

    def __init__(self, name, pickle_module=cloudpickle):
        super(PickleArtifact, self).__init__(name)

        self._obj = None

        if isinstance(pickle_module, str):
            self._pickle = __import__(pickle_module)
        else:
            self._pickle = pickle_module

    def _pkl_file_path(self, base_path):
        return os.path.join(base_path, self.name + PICKLE_FILE_EXTENSION)

    def pack(self, obj, metadata=None):  # pylint:disable=arguments-renamed
        self._obj = obj
        return self

    def load(self, path):
        with open(self._pkl_file_path(path), "rb") as pkl_file:
            obj = self._pickle.load(pkl_file)
        self.pack(obj)
        return self

    def get(self):
        return self._obj

    def save(self, dst):
        with open(self._pkl_file_path(dst), "wb") as pkl_file:
            self._pickle.dump(self._obj, pkl_file)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
from typing import Iterable, Sequence

from bentoml.adapters.base_output import BaseOutputAdapter, regroup_return_value
from bentoml.adapters.utils import NumpyJsonEncoder
from bentoml.types import (
    AwsLambdaEvent,
    HTTPResponse,
    InferenceError,
    InferenceResult,
    InferenceTask,
    JsonSerializable,
)

ApiFuncReturnValue = Sequence[JsonSerializable]


class JsonOutput(BaseOutputAdapter):
    """
    OutputAdapter converts returns of user defined API function into specific output,
    such as HTTP response, command line stdout or AWS Lambda event object.

    Args:
        cors (str): DEPRECATED. Moved to the configuration file.
            The value of the Access-Control-Allow-Origin header set in the
            HTTP/AWS Lambda response object. If set to None, the header will not be set.
            Default is None.
        ensure_ascii(bool): Escape all non-ASCII characters. Default False.
    """

    def __init__(self, ensure_ascii=False, **kwargs):
        super().__init__(**kwargs)
        self.ensure_ascii = ensure_ascii

    def pack_user_func_return_value(
        self, return_result: ApiFuncReturnValue, tasks: Sequence[InferenceTask],
    ) -> Sequence[InferenceResult[str]]:
        results = []
        for json_obj, task in regroup_return_value(return_result, tasks):
            args = task.cli_args
            if args:
                parser = argparse.ArgumentParser()
                parser.add_argument(
                    "-o", "--output", default="json", choices=["str", "json"]
                )
                parsed_args, _ = parser.parse_known_args(args)
                output = parsed_args.output
            else:
                output = "json"
            try:
                if output == "json":
                    json_str = json.dumps(
                        json_obj, cls=NumpyJsonEncoder, ensure_ascii=self.ensure_ascii
                    )
                else:
                    json_str = str(json_obj)
                results.append(
                    InferenceResult(
                        data=json_str,
                        http_status=200,
                        http_headers={"Content-Type": "application/json"},
                    )
                )
            except AssertionError as e:
                results.append(InferenceError(err_msg=str(e), http_status=400,))
            except Exception as e:  # pylint: disable=broad-except
                results.append(InferenceError(err_msg=str(e), http_status=500,))
        return tuple(results)

    def to_http_response(self, result: InferenceResult) -> HTTPResponse:
        return HTTPResponse.new(
            status=result.http_status,
            headers=result.http_headers,
            body=result.err_msg or result.data,
        )

    def to_cli(self, results: Iterable[InferenceResult]) -> int:
        """Handles an CLI command call, convert CLI arguments into
        corresponding data format that user API function is expecting, and
        prints the API function result to console output

        :param args: CLI arguments
        """
        flag = 0
        for result in results:
            if result.err_msg:
                print(result.err_msg)
                flag = 1
            else:
                print(result.data)
        return flag

    def to_aws_lambda_event(self, result: InferenceResult) -> AwsLambdaEvent:
        return {
            "statusCode": result.http_status,
            "body": result.err_msg or result.data,
        }



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import lru_cache
import logging
import os

from bentoml import __version__, _version as version_mod


# Note this file is loaded prior to logging being configured, thus logger is only
# used within functions in this file
logger = logging.getLogger(__name__)


DEBUG_ENV_VAR = "BENTOML_DEBUG"


def expand_env_var(env_var):
    """Expands potentially nested env var by repeatedly applying `expandvars` and
    `expanduser` until interpolation stops having any effect.
    """
    if not env_var:
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


# This is used as default for config('core', 'bentoml_deploy_version') - which is used
# for getting the BentoML PyPI version string or the URL to a BentoML sdist, indicating
# the BentoML module to be used when loading and using a saved BentoService bundle.
# This is useful when using customized BentoML fork/branch or when working with
# development branches of BentoML
BENTOML_VERSION = __version__
# e.g. from '0.4.2+5.g6cac97f.dirty' to '0.4.2'
LAST_PYPI_RELEASE_VERSION = __version__.split('+')[0]


def _is_pip_installed_bentoml():
    is_installed_package = hasattr(version_mod, 'version_json')
    is_tagged = not __version__.startswith('0+untagged')
    is_clean = not version_mod.get_versions()['dirty']
    return is_installed_package and is_tagged and is_clean


def get_local_config_file():
    if "BENTOML_CONFIG" in os.environ:
        # User local config file for customizing bentoml
        return expand_env_var(os.environ.get("BENTOML_CONFIG"))
    return None


@lru_cache(maxsize=1)
def get_bentoml_deploy_version(bentoml_deploy_version: str):
    """
    BentoML version to use for generated docker image or serverless function bundle to
    be deployed, this can be changed to an url to your fork of BentoML on github, or an
    url to your custom BentoML build, for example:

    bentoml_deploy_version = git+https://github.com/{username}/bentoml.git@{branch}
    """

    if bentoml_deploy_version != LAST_PYPI_RELEASE_VERSION:
        logger.info(f"Setting BentoML deploy version to '{bentoml_deploy_version}'")

    if LAST_PYPI_RELEASE_VERSION != BENTOML_VERSION:
        if _is_pip_installed_bentoml():
            logger.warning(
                "Using BentoML not from official PyPI release. In order to find the "
                "same version of BentoML when deploying your BentoService, you must "
                "set the 'core/bentoml_deploy_version' config to a http/git location "
                "of your BentoML fork, e.g.: 'bentoml_deploy_version = "
                "git+https://github.com/{username}/bentoml.git@{branch}'"
            )
        else:
            logger.warning(
                "Using BentoML installed in `editable` model, the local BentoML "
                "repository including all code changes will be packaged together with "
                "saved bundle created, under the './bundled_pip_dependencies' "
                "directory of the saved bundle."
            )
    return bentoml_deploy_version


def set_debug_mode(enabled: bool):
    os.environ[DEBUG_ENV_VAR] = str(enabled)

    # reconfigure logging
    from bentoml.utils.log import configure_logging

    configure_logging()

    logger.debug(
        f"Setting debug mode: {'ON' if enabled else 'OFF'} for current session"
    )


def get_debug_mode():
    if DEBUG_ENV_VAR in os.environ:
        return os.environ[DEBUG_ENV_VAR].lower() == "true"
    return False


def inject_dependencies():
    """Inject dependencies and configuration to BentoML packages"""

    from timeit import default_timer as timer

    start = timer()

    logger.debug("Start dependency injection")

    from bentoml.configuration.containers import BentoMLContainer, BentoMLConfiguration

    config_file = get_local_config_file()
    if config_file and config_file.endswith(".yml"):
        configuration = BentoMLConfiguration(override_config_file=config_file)
    else:
        configuration = BentoMLConfiguration()

    container = BentoMLContainer
    container.config.set(configuration.as_dict())

    end = timer()

    logger.debug("Dependency injection completed in %.3f seconds", end - start)



import logging
import os
import shutil

from bentoml.exceptions import InvalidArgument, MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv

logger = logging.getLogger(__name__)


def _import_fastai1_module():
    try:
        import fastai.basic_train
    except ImportError:
        raise MissingDependencyException(
            "fastai package is required to use " "bentoml.artifacts.FastaiModelArtifact"
        )

    return fastai


def _import_fastai2_module():
    try:
        import fastai.basics  # noqa
    except ImportError:
        raise MissingDependencyException(
            "fastai2 package is required to use "
            "bentoml.artifacts.FastaiModelArtifact"
        )

    return fastai


class Fastai1ModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading FastAI v1 Model

    Args:
        name (str): Name for the fastai model

    Raises:
        MissingDependencyException: Require fastai package to use Fast ai model artifact
        InvalidArgument: invalid argument type, model being packed must be instance of
            fastai.basic_train.Learner

    Example usage:

    >>> from fastai.tabular import *
    >>>
    >>> # prepare data
    >>> data = TabularList.from_df(...)
    >>> learn = tabular_learner(data, ...)
    >>> # train model
    >>>
    >>> import bentoml
    >>> from bentoml.adapters import DataframeInput
    >>> from bentoml.frameworks.fastai import Fastai1ModelArtifact
    >>>
    >>> @bentoml.artifacts([Fastai1ModelArtifact('model')])
    >>> @bentoml.env(infer_pip_packages=True)
    >>> class FastaiModelService(bentoml.BentoService):
    >>>
    >>>     @api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         results = []
    >>>         for _, row in df.iterrows():
    >>>             prediction = self.artifacts.model.predict(row)
    >>>             results.append(prediction[0].obj)
    >>>         return results
    >>>
    >>> svc = FastaiModelService()
    >>>
    >>> # Pack fastai basic_learner directly
    >>> svc.pack('model', learn)
    """

    def __init__(self, name):
        super().__init__(name)
        self._file_name = name + '.pkl'
        self._model = None

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self._file_name)

    def pack(self, model, metadata=None):  # pylint:disable=arguments-differ
        fastai_module = _import_fastai1_module()

        if not isinstance(model, fastai_module.basic_train.Learner):
            raise InvalidArgument(
                "Expect `model` argument to be `fastai.basic_train.Learner` instance"
            )

        self._model = model
        return self

    def load(self, path):
        fastai_module = _import_fastai1_module()

        model = fastai_module.basic_train.load_learner(path, self._file_name)
        return self.pack(model)

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            logger.warning(
                "BentoML by default does not include spacy and torchvision package "
                "when using FastaiModelArtifact. To make sure BentoML bundle those "
                "packages if they are required for your model, either import those "
                "packages in BentoService definition file or manually add them via "
                "`@env(pip_packages=['torchvision'])` when defining a BentoService"
            )
            env.add_pip_packages(['torch', "fastai<2.0.0"])

    def save(self, dst):
        self._model.export(file=self._file_name)

        shutil.copyfile(
            os.path.join(self._model.path, self._file_name), self._model_file_path(dst),
        )

    def get(self):
        return self._model


class FastaiModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading FastAI v2 Model

    Args:
        name (str): Name for the fastai v2 model

    Raises:
        MissingDependencyException: Require fastai>=2.0.0 package for Fastai model
            artifact
        InvalidArgument: invalid argument type, model being packed must be instance of
            fastai.basics.Learner

    Example usage:

    >>> from fastai.vision.learner import cnn_learner
    >>>
    >>> learner = cnn_learner(...)
    >>> # train model
    >>> # learner.fit_one_cycle(1)
    >>>
    >>> import bentoml
    >>> from bentoml.frameworks.fastai import FastaiModelArtifact
    >>> from bentoml.adapters import FileAdapter
    >>>
    >>> @bentoml.artifact([FastaiModelArtifact('learner')])
    >>> @bentoml.env(infer_pip_packages=True)
    >>> class FastaiImageService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=FileAdapter(), batch=True)
    >>>     def predict(self, files):
    >>>         dl = self.artifacts.learner.dls.test_dl(files, ...)
    >>>         result = self.artifacts.learner.get_preds(dl=dl)
    >>>         return result
    >>>
    >>> svc = FastaiImageService()
    >>> svc.pack('learner', learner)
    """

    def __init__(self, name):
        super().__init__(name)
        self._file_name = name + '.pkl'
        self._model = None

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self._file_name)

    def pack(self, model):  # pylint:disable=arguments-differ
        fastai2_module = _import_fastai2_module()

        if not isinstance(model, fastai2_module.basics.Learner):
            raise InvalidArgument(
                "Expect `model` argument to be `fastai.basics.Learner` instance"
            )

        self._model = model
        return self

    def load(self, path):
        fastai2_module = _import_fastai2_module()

        model = fastai2_module.basics.load_learner(os.path.join(path, self._file_name))
        return self.pack(model)

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            logger.warning(
                "BentoML by default does not include spacy and torchvision package "
                "when using FastaiModelArtifact. To make sure BentoML bundle those "
                "packages if they are required for your model, either import those "
                "packages in BentoService definition file or manually add them via "
                "`@env(pip_packages=['torchvision'])` when defining a BentoService"
            )
            env.add_pip_packages(['torch', "fastcore", "fastai>=2.0.0"])

    def save(self, dst):
        self._model.export(fname=self._file_name)
        shutil.copyfile(
            os.path.join(self._model.path, self._file_name), self._model_file_path(dst),
        )

    def get(self):
        return self._model



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import click
from tabulate import tabulate

from bentoml.utils.lazy_loader import LazyLoader
from bentoml.cli.click_utils import (
    _echo,
    _is_valid_bento_tag,
    _is_valid_bento_name,
)
from bentoml.cli.utils import (
    human_friendly_age_from_datetime,
    _format_labels_for_print,
)
from bentoml.utils import pb_to_yaml
from bentoml.yatai.client import get_yatai_client

yatai_proto = LazyLoader('yatai_proto', globals(), 'bentoml.yatai.proto')


def _print_bento_info(bento, output_type):
    if output_type == 'yaml':
        _echo(pb_to_yaml(bento))
    else:
        from google.protobuf.json_format import MessageToJson

        _echo(MessageToJson(bento))


def parse_delete_targets_argument_callback(
    ctx, params, value
):  # pylint: disable=unused-argument
    if value is None:
        return value
    delete_targets = value.split(",")
    delete_targets = list(map(str.strip, delete_targets))
    for delete_target in delete_targets:
        if not (
            _is_valid_bento_tag(delete_target) or _is_valid_bento_name(delete_target)
        ):
            raise click.BadParameter(
                "Bad formatting. Please present a valid bento bundle name or "
                "\"name:version\" tag. For list of bento bundles, separate delete "
                "targets by \",\", for example: \"my_service:v1,my_service:v2,"
                "classifier\""
            )
    return delete_targets


def _print_bento_table(bentos, wide=False):
    table = []
    if wide:
        headers = ['BENTO_SERVICE', 'CREATED_AT', 'APIS', 'ARTIFACTS', 'LABELS', 'URI']
    else:
        headers = ['BENTO_SERVICE', 'AGE', 'APIS', 'ARTIFACTS', 'LABELS']

    for bento in bentos:
        artifacts = [
            f'{artifact.name}<{artifact.artifact_type}>'
            for artifact in bento.bento_service_metadata.artifacts
        ]
        apis = [
            f'{api.name}<{api.input_type}:{api.output_type}>'
            for api in bento.bento_service_metadata.apis
        ]
        if wide:
            created_at = bento.bento_service_metadata.created_at.ToDatetime().strftime(
                "%Y-%m-%d %H:%M"
            )
        else:
            created_at = human_friendly_age_from_datetime(
                bento.bento_service_metadata.created_at.ToDatetime()
            )
        row = [
            f'{bento.name}:{bento.version}',
            created_at,
            ', '.join(apis),
            ', '.join(artifacts),
            _format_labels_for_print(bento.bento_service_metadata.labels),
        ]
        if wide:
            row.append(bento.uri.uri)
        table.append(row)

    table_display = tabulate(table, headers, tablefmt='plain')
    _echo(table_display)


def _print_bentos_info(bentos, output_type):
    if output_type == 'table':
        _print_bento_table(bentos)
    elif output_type == 'wide':
        _print_bento_table(bentos, wide=True)
    else:
        for bento in bentos:
            _print_bento_info(bento, output_type)


def add_bento_sub_command(cli):
    # pylint: disable=unused-variable
    @cli.command(help='Get BentoService information')
    @click.argument('bento', type=click.STRING)
    @click.option(
        '--limit', type=click.INT, help='Limit how many resources will be retrieved'
    )
    @click.option('--ascending-order', is_flag=True)
    @click.option('--print-location', is_flag=True)
    @click.option(
        '--labels',
        type=click.STRING,
        help="Label query to filter BentoServices, supports '=', '!=', 'IN', 'NotIn', "
        "'Exists', and 'DoesNotExist'. (e.g. key1=value1, key2!=value2, key3 "
        "In (value3, value3a), key4 DoesNotExist)",
    )
    @click.option(
        '--yatai-url',
        type=click.STRING,
        help='Remote YataiService URL. Optional. '
        'Example: "--yatai-url http://localhost:50050"',
    )
    @click.option(
        '-o', '--output', type=click.Choice(['json', 'yaml', 'table', 'wide'])
    )
    def get(bento, limit, ascending_order, print_location, labels, yatai_url, output):
        yc = get_yatai_client(yatai_url)
        if ':' in bento:
            result = yc.repository.get(bento)
            if print_location:
                _echo(result.uri.uri)
            else:
                _print_bento_info(result, output)
        else:
            output = output or 'table'
            result = yc.repository.list(
                bento_name=bento,
                limit=limit,
                ascending_order=ascending_order,
                labels=labels,
            )
            _print_bentos_info(result, output)

    @cli.command(name='list', help='List BentoServices information')
    @click.option(
        '--limit', type=click.INT, help='Limit how many BentoServices will be retrieved'
    )
    @click.option(
        '--offset', type=click.INT, help='How many BentoServices will be skipped'
    )
    @click.option(
        '--labels',
        type=click.STRING,
        help="Label query to filter BentoServices, supports '=', '!=', 'IN', 'NotIn', "
        "'Exists', and 'DoesNotExist'. (e.g. key1=value1, key2!=value2, key3 "
        "In (value3, value3a), key4 DoesNotExist)",
    )
    @click.option(
        '--order-by', type=click.Choice(['created_at', 'name']), default='created_at',
    )
    @click.option('--ascending-order', is_flag=True)
    @click.option(
        '--yatai-url',
        type=click.STRING,
        help='Remote YataiService URL. Optional. '
        'Example: "--yatai-url http://localhost:50050"',
    )
    @click.option(
        '-o',
        '--output',
        type=click.Choice(['json', 'yaml', 'table', 'wide']),
        default='table',
    )
    def list_bentos(
        limit, offset, labels, order_by, ascending_order, yatai_url, output
    ):
        yc = get_yatai_client(yatai_url)
        result = yc.repository.list(
            limit=limit,
            offset=offset,
            labels=labels,
            order_by=order_by,
            ascending_order=ascending_order,
        )
        _print_bentos_info(result, output)

    @cli.command()
    @click.argument(
        'delete_targets',
        type=click.STRING,
        callback=parse_delete_targets_argument_callback,
        required=False,
    )
    @click.option('--all', is_flag=True, help='Set true to delete all bento bundles')
    @click.option(
        '--labels',
        type=click.STRING,
        help="Label query to filter bento bundles, supports '=', '!=', 'IN', 'NotIn', "
        "'Exists', and 'DoesNotExist'. (e.g. key1=value1, key2!=value2, key3 "
        "In (value3, value3a), key4 DoesNotExist)",
    )
    @click.option(
        '--yatai-url',
        type=click.STRING,
        help='Remote YataiService URL. Optional. Example: '
        '"--yatai-url http://localhost:50050"',
    )
    @click.option(
        '-y',
        '--yes',
        '--assume-yes',
        is_flag=True,
        help='Skip confirmation when deleting a specific bento bundle',
    )
    def delete(
        all,  # pylint: disable=redefined-builtin
        delete_targets,
        labels,
        yatai_url,
        yes,  # pylint: disable=redefined-builtin
    ):
        """Delete bento bundles in target YataiService. When the --yatai-url option is not specified, it will use local Yatai by default.

Specify target service bundles to remove:

* Delete single bento bundle by "name:version", e.g: `bentoml delete IrisClassifier:v1`

* Bulk delete all bento bundles with a specific name, e.g.: `bentoml delete IrisClassifier`

* Bulk delete multiple bento bundles by name and version, separated by ",", e.g.: `benotml delete Irisclassifier:v1,MyPredictService:v2`

* Bulk delete by tag, e.g.: `bentoml delete --tag env=dev`

* Bulk delete all, e.g.: `bentoml delete --all`
        """  # noqa
        yc = get_yatai_client(yatai_url)
        # Backward compatible with the previous CLI, allows deletion with tag/s
        if delete_targets is not None and len(delete_targets) > 0:
            for item in delete_targets:
                if ':' in item:
                    if not _is_valid_bento_tag(item):
                        raise click.BadParameter(
                            "Bad formatting. Please present in BentoName:Version, "
                            "for example \"iris_classifier:v1.2.0\". For list of "
                            "BentoService, separate tags by \",\", for example: "
                            "\"my_service:v1,my_service:v2,classifier:v3\""
                        )
                    yc.repository.delete(
                        prune=all,
                        labels=labels,
                        bento_tag=item,
                        require_confirm=False if yes else True,
                    )
                else:
                    yc.repository.delete(
                        prune=all,
                        labels=labels,
                        bento_name=item,
                        require_confirm=False if yes else True,
                    )
        else:
            yc.repository.delete(
                prune=all, labels=labels, require_confirm=False if yes else True,
            )

    @cli.command(help='Pull BentoService from remote yatai server',)
    @click.argument("bento", type=click.STRING)
    @click.option(
        '--yatai-url',
        required=True,
        help='Remote YataiService URL. Example: "--yatai-url http://localhost:50050"',
    )
    def pull(bento, yatai_url):
        if ':' not in bento:
            _echo(f'BentoService {bento} invalid - specify name:version')
            return
        yc = get_yatai_client(yatai_url)
        yc.repository.pull(bento=bento)
        _echo(f'Pulled {bento} from {yatai_url}')

    @cli.command(help='Push BentoService to remote yatai server')
    @click.argument("bento", type=click.STRING)
    @click.option(
        '--yatai-url',
        required=True,
        help='Remote YataiService URL. Example: "--yatai-url http://localhost:50050"',
    )
    @click.option(
        '--with-labels/--without-labels',
        default=True,
        help="Retain bento bundle's labels or not for push Bento bundle to remote "
        "yatai. When running with --without-labels, labels are not retained in the "
        "remote yatai server",
    )
    def push(bento, yatai_url, with_labels):
        if ':' not in bento:
            _echo(f'BentoService {bento} invalid - specify name:version')
            return
        yc = get_yatai_client(yatai_url)
        yc.repository.push(bento, with_labels)
        _echo(f'Pushed {bento} to {yatai_url}')

    @cli.command(help='Retrieve')
    @click.argument("bento", type=click.STRING)
    @click.option(
        '--yatai-url',
        help='Remote YataiService URL. Example: "--yatai-url http://localhost:50050"',
    )
    @click.option(
        '--target_dir',
        help="Target directory to save BentoService. Defaults to the current directory",
        default=os.getcwd(),
    )
    def retrieve(bento, yatai_url, target_dir):
        yc = get_yatai_client(yatai_url)
        bento_pb = yc.repository.get(bento)
        yc.repository.download_to_directory(bento_pb, target_dir)

        _echo(f'Save {bento} artifact to directory {target_dir}')



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from contextlib import contextmanager

from urllib.parse import urlparse

from bentoml.exceptions import BentoMLException, LockUnavailable
from bentoml.yatai.db.base import Base
from bentoml.yatai.db.stores.deployment import DeploymentStore
from bentoml.yatai.db.stores.label import LabelStore
from bentoml.yatai.db.stores.metadata import MetadataStore

logger = logging.getLogger(__name__)


def is_postgresql_db(db_url):
    try:
        return urlparse(db_url).scheme == 'postgresql'
    except ValueError:
        return False


def is_sqlite_db(db_url):
    try:
        return urlparse(db_url).scheme == 'sqlite'
    except ValueError:
        return False


class DB(object):
    def __init__(self, db_url):
        from sqlalchemy import create_engine
        from sqlalchemy_utils import database_exists
        from sqlalchemy.orm import sessionmaker

        extra_db_args = {'echo': True}

        self.db_url = db_url
        if is_sqlite_db(db_url):
            extra_db_args['connect_args'] = {'check_same_thread': False}
            extra_db_args['echo'] = False
        elif is_postgresql_db(db_url):
            extra_db_args['connect_args'] = {'application_name': 'yatai'}
            extra_db_args['pool_pre_ping'] = True

        self.engine = create_engine(db_url, **extra_db_args)

        if not database_exists(self.engine.url) and not is_sqlite_db(db_url):
            raise BentoMLException(
                f'Database does not exist or Database name is missing in config '
                f'db.url: {db_url}'
            )

        self.create_all_or_upgrade_db()
        self.session_maker = sessionmaker(bind=self.engine)
        self._setup_stores()

    def _setup_stores(self):
        self.deployment_store = DeploymentStore()
        self.metadata_store = MetadataStore()
        self.label_store = LabelStore()

    @contextmanager
    def create_session(self):
        session = self.session_maker()
        try:
            yield session
            session.commit()
        except LockUnavailable as e:
            # rollback if lock cannot be acquired, bubble error up
            session.rollback()
            raise LockUnavailable(e)
        except Exception as e:
            session.rollback()
            raise BentoMLException(e)
        finally:
            session.close()

    def create_all_or_upgrade_db(self):
        # alembic add a lot of import time, so we lazy import
        from alembic import command
        from alembic.config import Config
        from sqlalchemy import inspect

        alembic_config_file = os.path.join(os.path.dirname(__file__), '../alembic.ini')
        alembic_config = Config(alembic_config_file)
        alembic_config.set_main_option('sqlalchemy.url', self.db_url)

        inspector = inspect(self.engine)
        tables = inspector.get_table_names()

        if 'deployments' not in tables or 'bentos' not in tables:
            logger.debug('Creating tables')
            Base.metadata.create_all(self.engine)
            command.stamp(alembic_config, 'head')
        else:
            logger.debug('Upgrading tables to the latest revision')
            command.upgrade(alembic_config, 'heads')



import logging
from typing import Optional, Dict
from urllib.parse import urlparse

import docker

from bentoml.exceptions import MissingDependencyException, BentoMLException

logger = logging.getLogger(__name__)


def ensure_docker_available_or_raise() -> None:
    """
    Ensure docker is available.

    Raises:
        :class:`~MissingDependencyException`:
            for :class:`~docker.errors.APIErrors`
            or :class:`~docker.errors.DockerException`
    """
    try:
        client = docker.from_env()
        client.ping()
    except docker.errors.APIError as error:
        raise MissingDependencyException(f'Docker server is not responsive. {error}')
    except docker.errors.DockerException:
        raise MissingDependencyException(
            'Docker is required for this deployment. Please visit '
            'www.docker.com for instructions'
        )


def _strip_scheme(url: str) -> str:
    """
    Stripe url's schema.

    Examples:
    http://some.url/path -> some.url/path

    Args:
        url (`str`)

    Returns:
        :obj:`str`
    """
    parsed = urlparse(url)
    scheme = "%s://" % parsed.scheme
    return parsed.geturl().replace(scheme, "", 1)


def generate_docker_image_tag(
    image_name: str, version: str = 'latest', registry_url=None
):
    image_tag = f'{image_name}:{version}'.lower()
    if registry_url is not None:
        return _strip_scheme(f'{registry_url}/{image_tag}')
    else:
        return image_tag


def build_docker_image(
    context_path: str,
    image_tag: str,
    dockerfile: Optional[str] = 'Dockerfile',
    additional_build_args: Optional[Dict[str, str]] = None,
):

    docker_client = docker.from_env()
    try:
        docker_client.images.build(
            path=context_path,
            tag=image_tag,
            dockerfile=dockerfile,
            buildargs=additional_build_args,
        )
    except (docker.errors.APIError, docker.errors.BuildError) as error:
        logger.error(f'Failed to build docker image {image_tag}: {error}')
        raise BentoMLException(f'Failed to build docker image {image_tag}: {error}')


def push_docker_image_to_repository(
    repository: str,
    image_tag: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
):
    docker_client = docker.from_env()
    docker_push_kwags = {'repository': repository, 'tag': image_tag}
    if username is not None and password is not None:
        docker_push_kwags['auth_config'] = {'username': username, 'password': password}
    try:
        docker_client.images.push(**docker_push_kwags)
    except docker.errors.APIError as error:
        raise BentoMLException(f'Failed to push docker image {image_tag}: {error}')



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


def is_s3_url(url):
    """
    Check if url is an s3, s3n, or s3a url
    """
    try:
        return urlparse(url).scheme in ["s3", "s3n", "s3a"]
    except ValueError:
        return False


def create_s3_bucket_if_not_exists(bucket_name, region):
    import boto3
    from botocore.exceptions import ClientError

    s3_client = boto3.client('s3', region)
    try:
        s3_client.get_bucket_acl(Bucket=bucket_name)
        logger.debug("Found bucket %s in region %s already exist", bucket_name, region)
    except ClientError as error:
        if error.response and error.response['Error']['Code'] == 'NoSuchBucket':
            logger.debug('Creating s3 bucket: %s in region %s', bucket_name, region)

            # NOTE: boto3 will raise ClientError(InvalidLocationConstraint) if
            # `LocationConstraint` is set to `us-east-1` region.
            # https://github.com/boto/boto3/issues/125.
            # This issue still show up in  boto3 1.13.4(May 6th 2020)
            try:
                s3_client.create_bucket(
                    Bucket=bucket_name,
                    CreateBucketConfiguration={'LocationConstraint': region},
                )
            except ClientError as s3_error:
                if (
                    s3_error.response
                    and s3_error.response['Error']['Code']
                    == 'InvalidLocationConstraint'
                ):
                    logger.debug(
                        'Special s3 region: %s, will attempt create bucket without '
                        '`LocationConstraint`',
                        region,
                    )
                    s3_client.create_bucket(Bucket=bucket_name)
                else:
                    raise s3_error
        else:
            raise error



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import asyncio
import requests
from contextlib import contextmanager
from contextvars import ContextVar


trace_stack_var = ContextVar('trace_stack', default=None)


def _load_http_headers(headers):
    from py_zipkin.zipkin import ZipkinAttrs  # pylint: disable=E0401

    if not headers or "X-B3-TraceId" not in headers:
        return None

    return ZipkinAttrs(
        headers.get("X-B3-TraceId"),
        headers.get("X-B3-SpanId"),
        headers.get("X-B3-ParentSpanId"),
        headers.get("X-B3-Flags") or '0',
        False if headers.get("X-B3-Sampled") == '0' else True,
    )


def _set_http_headers(attrs, headers):
    if not headers or "X-B3-TraceId" in headers:
        return

    tracing_headers = {
        "X-B3-TraceId": attrs.trace_id,
        "X-B3-SpanId": attrs.span_id,
        "X-B3-Flags": attrs.flags,
        "X-B3-Sampled": attrs.is_sampled and '1' or '0',
    }
    if attrs.parent_span_id:
        tracing_headers["X-B3-ParentSpanId"] = attrs.parent_span_id
    headers.update(tracing_headers)


def _make_child_attrs(attrs):
    from py_zipkin.zipkin import ZipkinAttrs  # pylint: disable=E0401
    from py_zipkin.util import generate_random_64bit_string  # pylint: disable=E0401

    return ZipkinAttrs(
        attrs.trace_id,
        generate_random_64bit_string(),
        attrs.span_id,
        attrs.flags,
        attrs.is_sampled,
    )


def _make_new_attrs(sample_rate=1.0):
    from py_zipkin.zipkin import ZipkinAttrs  # pylint: disable=E0401
    from py_zipkin.util import generate_random_64bit_string  # pylint: disable=E0401

    return ZipkinAttrs(
        generate_random_64bit_string(),
        generate_random_64bit_string(),
        None,
        '0',
        sample_rate and random.random() < sample_rate or False,
    )


def get_zipkin_tracer(server_url):
    from py_zipkin.transport import BaseTransportHandler  # pylint: disable=E0401

    class HttpTransport(BaseTransportHandler):
        def __init__(self, server_url):
            super(HttpTransport, self).__init__()
            self.server_url = server_url

        def get_max_payload_bytes(self):
            # None for no max payload size
            return None

        def send(self, payload):
            requests.post(
                self.server_url,
                data=payload,
                headers={'Content-Type': 'application/x-thrift'},
            )

    class AsyncHttpTransport(BaseTransportHandler):
        '''
        add trace data transporting task into default eventloop
        '''

        def __init__(self, server_url):
            super(AsyncHttpTransport, self).__init__()
            self.server_url = server_url

        def get_max_payload_bytes(self):
            # None for no max payload size
            return None

        @staticmethod
        async def _async_post(url, data, headers):
            from aiohttp import ClientSession

            async with ClientSession() as client:
                async with client.post(url, data=data, headers=headers) as resp:
                    resp = await resp.text()
                    return resp

        def send(self, payload):
            asyncio.get_event_loop().create_task(
                self._async_post(
                    self.server_url,
                    data=payload,
                    headers={'Content-Type': 'application/x-thrift'},
                )
            )

    class ZipkinTracer:
        def __init__(self, server_url):
            self.server_url = server_url
            self.async_transport = AsyncHttpTransport(self.server_url)
            self.http_transport = HttpTransport(self.server_url)

        @contextmanager
        def span(
            self,
            service_name,
            span_name,
            request_headers=None,
            async_transport=False,
            sample_rate=1.0,
            standalone=False,
            is_root=False,
        ):
            from py_zipkin import Encoding  # pylint: disable=E0401
            from py_zipkin.zipkin import zipkin_span  # pylint: disable=E0401

            trace_stack = trace_stack_var.get()

            parent_attrs = _load_http_headers(request_headers) or trace_stack or None

            if not is_root and parent_attrs:
                attrs = _make_child_attrs(parent_attrs)
            else:
                attrs = _make_new_attrs(sample_rate)

            if not attrs.is_sampled:
                if standalone:
                    yield None
                    return
                else:
                    token = trace_stack_var.set(attrs)
                    _set_http_headers(attrs, request_headers)
                    yield attrs
                    trace_stack_var.reset(token)
                    return

            if async_transport:
                transport_handler = self.async_transport
            else:
                transport_handler = self.http_transport

            with zipkin_span(
                service_name=service_name,
                span_name=span_name,
                zipkin_attrs=attrs,
                transport_handler=transport_handler,
                encoding=Encoding.V2_JSON,
            ):
                if standalone:
                    yield
                    return
                else:
                    token = trace_stack_var.set(attrs)
                    _set_http_headers(attrs, request_headers)
                    yield
                    trace_stack_var.reset(token)

        @contextmanager
        def async_span(self, *args, **kwargs):
            with self.span(*args, async_transport=True, **kwargs) as ctx:
                yield ctx
            return

    return ZipkinTracer(server_url)



import logging
import os
import pathlib
import shutil

from bentoml.exceptions import (
    BentoMLException,
    InvalidArgument,
    MissingDependencyException,
)
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv

logger = logging.getLogger(__name__)


SUPPORTED_ONNX_BACKEND = ["onnxruntime", "onnxruntime-gpu"]


def _is_path_like(path):
    return isinstance(path, (str, bytes, pathlib.Path, os.PathLike))


def _is_onnx_model_file(path):
    return (
        _is_path_like(path)
        and os.path.isfile(path)
        and str(path).lower().endswith(".onnx")
    )


class OnnxModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading ONNX Model

    Args:
        name (string): Name of the artifact
        backend (string): Name of ONNX inference runtime. ["onnxruntime",
                            "onnxruntime-gpu"]

    Raises:
        MissingDependencyException: onnx package is required for packing a ModelProto
                                    object
        NotImplementedError: {backend} as onnx runtime is not supported at the moment

    Example usage:

    >>> # Train a model.
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> iris = load_iris()
    >>> X, y = iris.data, iris.target
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>> clr = RandomForestClassifier()
    >>> clr.fit(X_train, y_train)
    >>> # Convert into ONNX format
    >>> from skl2onnx import convert_sklearn
    >>> from skl2onnx.common.data_types import FloatTensorType
    >>> initial_type = [('float_input', FloatTensorType([None, 4]))]
    >>>
    >>> onnx_model = convert_sklearn(clr, initial_types=initial_type)
    >>> with open("rf_iris.onnx", "wb") as f:
    >>>     f.write(onnx_model.SerializeToString())
    >>>
    >>>
    >>> import numpy
    >>> import bentoml
    >>> from bentoml.frameworks.onnx import OnnxModelArtifact
    >>> from bentoml.adapters import DataframeInput
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([OnnxModelArtifact('model', backend='onnxruntime')])
    >>> class OnnxIrisClassifierService(bentoml.BentoService):
    >>>     @bentoml.api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         input_data = df.to_numpy().astype(numpy.float32
    >>>         input_name = self.artifacts.model.get_inputs()[0].name
    >>>         output_name = self.artifacts.model.get_outputs()[0].name
    >>>         return self.artifacts.model.run(
    >>>                     [output_name], {input_name: input_data}
    >>>                )[0]
    >>>
    >>> svc = OnnxIrisClassifierService()
    >>>
    >>> # Option one: pack with path to model on local system
    >>> svc.pack('model', './rf_iris.onnx')
    >>>
    >>> # Option two: pack with ONNX model object
    >>> # svc.pack('model', onnx_model)
    >>>
    >>> # Save BentoService
    >>> svc.save()
    """

    def __init__(self, name, backend="onnxruntime"):
        super().__init__(name)
        if backend not in SUPPORTED_ONNX_BACKEND:
            raise BentoMLException(
                f'"{backend}" runtime is currently not supported for OnnxModelArtifact'
            )
        self.backend = backend
        self._inference_session = None
        self._onnx_model_path = None
        self._model_proto = None

    def _saved_model_file_path(self, base_path):
        return os.path.join(base_path, self.name + ".onnx")

    def pack(
        self, path_or_model_proto, metadata=None
    ):  # pylint:disable=arguments-renamed
        if _is_onnx_model_file(path_or_model_proto):
            self._onnx_model_path = path_or_model_proto
        else:
            try:
                import onnx

                if isinstance(path_or_model_proto, onnx.ModelProto):
                    self._model_proto = path_or_model_proto
                else:
                    raise InvalidArgument(
                        "onnx.ModelProto or a .onnx model file path is required to "
                        "pack an OnnxModelArtifact"
                    )
            except ImportError:
                raise InvalidArgument(
                    "onnx.ModelProto or a .onnx model file path is required to pack "
                    "an OnnxModelArtifact"
                )

        assert self._onnx_model_path or self._model_proto, (
            "Either self._onnx_model_path or self._model_proto has to be initialized "
            "after initializing _OnnxModelArtifactWrapper"
        )

        return self

    def load(self, path):
        return self.pack(self._saved_model_file_path(path))

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            if self.backend == "onnxruntime":
                env.add_pip_packages(["onnxruntime"])
            elif self.backend == "onnxruntime-gpu":
                env.add_pip_packages(["onnxruntime-gpu"])

    def _get_onnx_inference_session(self):
        if self.backend == "onnxruntime" or self.backend == "onnxruntime-gpu":
            try:
                import onnxruntime
            except ImportError:
                raise MissingDependencyException(
                    f'"{self.backend}" package is required for inference with '
                    f'"{self.backend}" as backend"'
                )

            if self._model_proto:
                logger.info(
                    "Initializing onnxruntime InferenceSession with onnx.ModelProto "
                    "instance"
                )
                return onnxruntime.InferenceSession(
                    self._model_proto.SerializeToString()
                )
            elif self._onnx_model_path:
                logger.info(
                    "Initializing onnxruntime InferenceSession from onnx file:"
                    f"'{self._onnx_model_path}'"
                )
                return onnxruntime.InferenceSession(self._onnx_model_path)
            else:
                raise BentoMLException("OnnxModelArtifact in bad state")
        else:
            raise BentoMLException(
                f'"{self.backend}" runtime is currently not supported for '
                f"OnnxModelArtifact"
            )

    def get(self):
        if not self._inference_session:
            self._inference_session = self._get_onnx_inference_session()
        return self._inference_session

    def save(self, dst):
        if self._onnx_model_path:
            shutil.copyfile(self._onnx_model_path, self._saved_model_file_path(dst))
        elif self._model_proto:
            try:
                import onnx
            except ImportError:
                raise MissingDependencyException(
                    '"onnx" package is required for packing with OnnxModelArtifact'
                )
            onnx.save_model(self._model_proto, self._saved_model_file_path(dst))
        else:
            raise InvalidArgument(
                "onnx.ModelProto or a model file path is required to pack an "
                "OnnxModelArtifact"
            )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bentoml.adapters.annotated_image_input import AnnotatedImageInput
from bentoml.adapters.base_input import BaseInputAdapter
from bentoml.adapters.base_output import BaseOutputAdapter
from bentoml.adapters.clipper_input import (
    ClipperBytesInput,
    ClipperDoublesInput,
    ClipperFloatsInput,
    ClipperIntsInput,
    ClipperStringsInput,
)
from bentoml.adapters.dataframe_input import DataframeInput
from bentoml.adapters.dataframe_output import DataframeOutput
from bentoml.adapters.default_output import DefaultOutput
from bentoml.adapters.file_input import FileInput
from bentoml.adapters.image_input import ImageInput
from bentoml.adapters.json_input import JsonInput
from bentoml.adapters.json_output import JsonOutput
from bentoml.adapters.multi_file_input import MultiFileInput
from bentoml.adapters.multi_image_input import MultiImageInput
from bentoml.adapters.string_input import StringInput
from bentoml.adapters.tensorflow_tensor_input import TfTensorInput
from bentoml.adapters.tensorflow_tensor_output import TfTensorOutput

__all__ = [
    "BaseInputAdapter",
    'BaseOutputAdapter',
    "DataframeInput",
    'DataframeOutput',
    "TfTensorInput",
    'TfTensorOutput',
    "JsonInput",
    "StringInput",
    'JsonOutput',
    "ImageInput",
    "MultiImageInput",
    "FileInput",
    "MultiFileInput",
    "AnnotatedImageInput",
    "ClipperBytesInput",
    "ClipperDoublesInput",
    "ClipperFloatsInput",
    "ClipperIntsInput",
    "ClipperStringsInput",
    'DefaultOutput',
]



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Iterable, Iterator, Sequence, Tuple

from bentoml.exceptions import APIDeprecated
from bentoml.types import (
    ApiFuncReturnValue,
    AwsLambdaEvent,
    HTTPResponse,
    InferenceResult,
    InferenceTask,
    Output,
)


def regroup_return_value(
    return_values: Iterable, tasks: Sequence[InferenceTask]
) -> Iterator[Tuple[Output, InferenceTask]]:
    iter_rv = iter(return_values)
    try:
        for task in tasks:
            if task.batch is None:
                yield next(iter_rv), task
            else:
                yield tuple(itertools.islice(iter_rv, task.batch)), task
    except StopIteration:
        for task in tasks:
            task.discard(
                http_status=500,
                err_msg="The return values of Api Function doesn't match tasks",
            )


class BaseOutputAdapter:
    """
    Output adapters converts returns of user defined API function into specific output,
    such as HTTP response, command line stdout or AWS Lambda event object.

    Args:
        cors (str): DEPRECATED. Moved to the configuration file.
            The value of the Access-Control-Allow-Origin header set in the
            HTTP/AWS Lambda response object. If set to None, the header will not be set.
            Default is None.
    """

    def __init__(self, cors=None):
        if cors is not None:
            raise APIDeprecated(
                "setting cors from OutputAdapter is no more supported."
                "See cors option in the configuration file."
            )

    @property
    def config(self):
        return dict()

    @property
    def pip_dependencies(self):
        """
        :return: List of PyPI package names required by this OutputAdapter
        """
        return []

    def pack_user_func_return_value(
        self, return_result: ApiFuncReturnValue, tasks: Sequence[InferenceTask],
    ) -> Sequence[InferenceResult]:
        """
        Pack the return value of user defined API function into InferenceResults
        """
        raise NotImplementedError()

    def to_http_response(self, result: InferenceResult) -> HTTPResponse:
        """
        Converts InferenceResults into HTTP responses.
        """
        raise NotImplementedError()

    def to_cli(self, results: Iterable[InferenceResult]) -> int:
        """
        Converts InferenceResults into CLI output.
        """
        raise NotImplementedError()

    def to_aws_lambda_event(self, result: InferenceResult) -> AwsLambdaEvent:
        """
        Converts InferenceResults into AWS lambda events.
        """
        raise NotImplementedError()



import os
import base64
import json
import logging

from bentoml.utils.tempdir import TempDirectory

from bentoml.utils.lazy_loader import LazyLoader
from bentoml.saved_bundle import loader
from bentoml.yatai.deployment.aws_ec2.templates import (
    EC2_CLOUDFORMATION_TEMPLATE,
    EC2_USER_INIT_SCRIPT,
)
from bentoml.yatai.deployment.operator import DeploymentOperatorBase
from bentoml.yatai.proto.deployment_pb2 import (
    ApplyDeploymentResponse,
    DeleteDeploymentResponse,
    DescribeDeploymentResponse,
    DeploymentState,
)
from bentoml.yatai.status import Status
from bentoml.utils import status_pb_to_error_code_and_message
from bentoml.utils.s3 import create_s3_bucket_if_not_exists
from bentoml.yatai.deployment.docker_utils import (
    ensure_docker_available_or_raise,
    generate_docker_image_tag,
    build_docker_image,
    push_docker_image_to_repository,
)
from bentoml.yatai.deployment.aws_utils import (
    generate_aws_compatible_string,
    get_default_aws_region,
    ensure_sam_available_or_raise,
    validate_sam_template,
    FAILED_CLOUDFORMATION_STACK_STATUS,
    cleanup_s3_bucket_if_exist,
    delete_cloudformation_stack,
    delete_ecr_repository,
    get_instance_ip_from_scaling_group,
    get_aws_user_id,
    create_ecr_repository_if_not_exists,
    get_ecr_login_info,
    describe_cloudformation_stack,
)
from bentoml.exceptions import (
    BentoMLException,
    InvalidArgument,
    YataiDeploymentException,
)
from bentoml.yatai.proto.repository_pb2 import GetBentoRequest, BentoUri
from bentoml.yatai.proto import status_pb2
from bentoml.yatai.deployment.aws_ec2.utils import (
    build_template,
    package_template,
    deploy_template,
    get_endpoints_from_instance_address,
    get_healthy_target,
)
from bentoml.yatai.deployment.aws_ec2.constants import (
    BENTOSERVICE_PORT,
    TARGET_HEALTH_CHECK_INTERVAL,
    TARGET_HEALTH_CHECK_PATH,
    TARGET_HEALTH_CHECK_PORT,
    TARGET_HEALTH_CHECK_TIMEOUT_SECONDS,
    TARGET_HEALTH_CHECK_THRESHOLD_COUNT,
)

logger = logging.getLogger(__name__)

yatai_proto = LazyLoader("yatai_proto", globals(), "bentoml.yatai.proto")
SAM_TEMPLATE_NAME = "template.yml"


def _make_user_data(registry, tag, region):
    """
    Create init script for EC2 containers to download docker image,
    and run container on startup.
    args:
        registry: ECR registry domain
        tag: bento tag
        region: AWS region
    """

    base_format = EC2_USER_INIT_SCRIPT.format(
        registry=registry, tag=tag, region=region, bentoservice_port=BENTOSERVICE_PORT
    )
    encoded = base64.b64encode(base_format.encode("ascii")).decode("ascii")
    return encoded


def _make_cloudformation_template(
    project_dir,
    user_data,
    s3_bucket_name,
    sam_template_name,
    elb_name,
    ami_id,
    instance_type,
    autoscaling_min_size,
    autoscaling_desired_capacity,
    autoscaling_max_size,
):
    """
    Create and save cloudformation template for deployment
    args:
        project_dir: path to save template file
        user_data: base64 encoded user data for cloud-init script
        s3_bucket_name: AWS S3 bucket name
        sam_template_name: template name to save
        ami_id: ami id for EC2 container to use
        instance_type: EC2 instance type
        autocaling_min_size: autoscaling group minimum size
        autocaling_desired_capacity: autoscaling group desired size
        autocaling_min_size: autoscaling group maximum size


    NOTE: SSH ACCESS TO INSTANCE MAY NOT BE REQUIRED
    TODO: Port taken from cli
    """

    template_file_path = os.path.join(project_dir, sam_template_name)
    with open(template_file_path, "a") as f:
        f.write(
            EC2_CLOUDFORMATION_TEMPLATE.format(
                ami_id=ami_id,
                template_name=sam_template_name,
                instance_type=instance_type,
                user_data=user_data,
                elb_name=elb_name,
                autoscaling_min_size=autoscaling_min_size,
                autoscaling_desired_capacity=autoscaling_desired_capacity,
                autoscaling_max_size=autoscaling_max_size,
                s3_bucket_name=s3_bucket_name,
                target_health_check_interval_seconds=TARGET_HEALTH_CHECK_INTERVAL,
                target_health_check_path=TARGET_HEALTH_CHECK_PATH,
                target_health_check_port=TARGET_HEALTH_CHECK_PORT,
                target_health_check_timeout_seconds=TARGET_HEALTH_CHECK_TIMEOUT_SECONDS,
                target_health_check_threshold_count=TARGET_HEALTH_CHECK_THRESHOLD_COUNT,
            )
        )
    return template_file_path


def generate_ec2_resource_names(namespace, name):
    sam_template_name = generate_aws_compatible_string(
        f"btml-template-{namespace}-{name}"
    )
    deployment_stack_name = generate_aws_compatible_string(
        f"btml-stack-{namespace}-{name}"
    )
    repo_name = generate_aws_compatible_string(f"btml-repo-{namespace}-{name}")
    elb_name = generate_aws_compatible_string(f"{namespace}-{name}", max_length=32)

    return sam_template_name, deployment_stack_name, repo_name, elb_name


def deploy_ec2_service(
    deployment_pb,
    deployment_spec,
    bento_path,
    aws_ec2_deployment_config,
    s3_bucket_name,
    region,
):
    (
        sam_template_name,
        deployment_stack_name,
        repo_name,
        elb_name,
    ) = generate_ec2_resource_names(deployment_pb.namespace, deployment_pb.name)

    with TempDirectory() as project_path:
        repository_id = create_ecr_repository_if_not_exists(region, repo_name)
        registry_url, username, password = get_ecr_login_info(region, repository_id)
        ecr_tag = generate_docker_image_tag(
            repo_name, deployment_spec.bento_version, registry_url
        )
        build_docker_image(context_path=bento_path, image_tag=ecr_tag)
        push_docker_image_to_repository(
            repository=ecr_tag, username=username, password=password
        )

        logger.info("Generating user data")
        encoded_user_data = _make_user_data(registry_url, ecr_tag, region)

        logger.info("Making template")
        template_file_path = _make_cloudformation_template(
            project_path,
            encoded_user_data,
            s3_bucket_name,
            sam_template_name,
            elb_name,
            aws_ec2_deployment_config.ami_id,
            aws_ec2_deployment_config.instance_type,
            aws_ec2_deployment_config.autoscale_min_size,
            aws_ec2_deployment_config.autoscale_desired_capacity,
            aws_ec2_deployment_config.autoscale_max_size,
        )
        validate_sam_template(
            sam_template_name, aws_ec2_deployment_config.region, project_path
        )

        logger.info("Building service")
        build_template(
            template_file_path, project_path, aws_ec2_deployment_config.region
        )

        logger.info("Packaging service")
        package_template(s3_bucket_name, project_path, aws_ec2_deployment_config.region)

        logger.info("Deploying service")
        deploy_template(
            deployment_stack_name,
            s3_bucket_name,
            project_path,
            aws_ec2_deployment_config.region,
        )


class AwsEc2DeploymentOperator(DeploymentOperatorBase):
    def add(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            deployment_spec.aws_ec2_operator_config.region = (
                deployment_spec.aws_ec2_operator_config.region
                or get_default_aws_region()
            )
            if not deployment_spec.aws_ec2_operator_config.region:
                raise InvalidArgument("AWS region is missing")

            ensure_sam_available_or_raise()
            ensure_docker_available_or_raise()

            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )

            if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
                raise BentoMLException(
                    "BentoML currently not support {} repository".format(
                        BentoUri.StorageType.Name(bento_pb.bento.uri.type)
                    )
                )
            bento_path = bento_pb.bento.uri.uri

            return self._add(deployment_pb, bento_pb, bento_path)
        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = f"Error: {str(error)}"
            return ApplyDeploymentResponse(
                status=error.status_proto, deployment=deployment_pb
            )

    def _add(self, deployment_pb, bento_pb, bento_path):
        try:
            if loader._is_remote_path(bento_path):
                with loader._resolve_remote_bundle_path(bento_path) as local_path:
                    return self._add(deployment_pb, bento_pb, local_path)

            deployment_spec = deployment_pb.spec
            aws_ec2_deployment_config = deployment_spec.aws_ec2_operator_config

            user_id = get_aws_user_id()
            artifact_s3_bucket_name = generate_aws_compatible_string(
                "btml-{user_id}-{namespace}".format(
                    user_id=user_id, namespace=deployment_pb.namespace,
                )
            )
            create_s3_bucket_if_not_exists(
                artifact_s3_bucket_name, aws_ec2_deployment_config.region
            )
            deploy_ec2_service(
                deployment_pb,
                deployment_spec,
                bento_path,
                aws_ec2_deployment_config,
                artifact_s3_bucket_name,
                aws_ec2_deployment_config.region,
            )
        except BentoMLException as error:
            if artifact_s3_bucket_name and aws_ec2_deployment_config.region:
                cleanup_s3_bucket_if_exist(
                    artifact_s3_bucket_name, aws_ec2_deployment_config.region
                )
            raise error
        return ApplyDeploymentResponse(status=Status.OK(), deployment=deployment_pb)

    def delete(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            ec2_deployment_config = deployment_spec.aws_ec2_operator_config
            ec2_deployment_config.region = (
                ec2_deployment_config.region or get_default_aws_region()
            )
            if not ec2_deployment_config.region:
                raise InvalidArgument("AWS region is missing")

            _, deployment_stack_name, repository_name, _ = generate_ec2_resource_names(
                deployment_pb.namespace, deployment_pb.name
            )
            # delete stack
            delete_cloudformation_stack(
                deployment_stack_name, ec2_deployment_config.region
            )

            # delete repo from ecr
            delete_ecr_repository(repository_name, ec2_deployment_config.region)

            # remove bucket
            if deployment_pb.state.info_json:
                deployment_info_json = json.loads(deployment_pb.state.info_json)
                bucket_name = deployment_info_json.get('S3Bucket')
                if bucket_name:
                    cleanup_s3_bucket_if_exist(
                        bucket_name, ec2_deployment_config.region
                    )

            return DeleteDeploymentResponse(status=Status.OK())
        except BentoMLException as error:
            return DeleteDeploymentResponse(status=error.status_proto)

    def update(self, deployment_pb, previous_deployment):
        try:
            ensure_sam_available_or_raise()
            ensure_docker_available_or_raise()
            deployment_spec = deployment_pb.spec
            ec2_deployment_config = deployment_spec.aws_ec2_operator_config
            ec2_deployment_config.region = (
                ec2_deployment_config.region or get_default_aws_region()
            )
            if not ec2_deployment_config.region:
                raise InvalidArgument("AWS region is missing")

            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )

            if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
                raise BentoMLException(
                    "BentoML currently not support {} repository".format(
                        BentoUri.StorageType.Name(bento_pb.bento.uri.type)
                    )
                )

            return self._update(
                deployment_pb,
                previous_deployment,
                bento_pb.bento.uri.uri,
                ec2_deployment_config.region,
            )
        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = f"Error: {str(error)}"
            return ApplyDeploymentResponse(
                status=error.status_proto, deployment=deployment_pb
            )

    def _update(self, deployment_pb, previous_deployment_pb, bento_path, region):
        if loader._is_remote_path(bento_path):
            with loader._resolve_remote_bundle_path(bento_path) as local_path:
                return self._update(
                    deployment_pb, previous_deployment_pb, local_path, region
                )

        updated_deployment_spec = deployment_pb.spec
        updated_deployment_config = updated_deployment_spec.aws_ec2_operator_config

        describe_result = self.describe(deployment_pb)
        if describe_result.status.status_code != status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                describe_result.status
            )
            raise YataiDeploymentException(
                f"Failed fetching ec2 deployment current status - "
                f"{error_code}:{error_message}"
            )

        previous_deployment_state = json.loads(describe_result.state.info_json)
        if "S3Bucket" in previous_deployment_state:
            s3_bucket_name = previous_deployment_state.get("S3Bucket")
        else:
            raise BentoMLException(
                "S3 Bucket is missing in the AWS EC2 deployment, please make sure "
                "it exists and try again"
            )

        deploy_ec2_service(
            deployment_pb,
            updated_deployment_spec,
            bento_path,
            updated_deployment_config,
            s3_bucket_name,
            region,
        )

        return ApplyDeploymentResponse(status=Status.OK(), deployment=deployment_pb)

    def describe(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            ec2_deployment_config = deployment_spec.aws_ec2_operator_config
            ec2_deployment_config.region = (
                ec2_deployment_config.region or get_default_aws_region()
            )
            if not ec2_deployment_config.region:
                raise InvalidArgument("AWS region is missing")

            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )
            bento_service_metadata = bento_pb.bento.bento_service_metadata
            api_names = [api.name for api in bento_service_metadata.apis]

            _, deployment_stack_name, _, _ = generate_ec2_resource_names(
                deployment_pb.namespace, deployment_pb.name
            )
            try:
                stack_result = describe_cloudformation_stack(
                    ec2_deployment_config.region, deployment_stack_name
                )
                if stack_result.get("Outputs"):
                    outputs = stack_result.get("Outputs")
                else:
                    return DescribeDeploymentResponse(
                        status=Status.ABORTED('"Outputs" field is not present'),
                        state=DeploymentState(
                            state=DeploymentState.ERROR,
                            error_message='"Outputs" field is not present',
                        ),
                    )

                if stack_result["StackStatus"] in FAILED_CLOUDFORMATION_STACK_STATUS:
                    state = DeploymentState(state=DeploymentState.FAILED)
                    return DescribeDeploymentResponse(status=Status.OK(), state=state)

            except Exception as error:  # pylint: disable=broad-except
                state = DeploymentState(
                    state=DeploymentState.ERROR, error_message=str(error)
                )
                return DescribeDeploymentResponse(
                    status=Status.INTERNAL(str(error)), state=state
                )

            info_json = {}
            outputs = {o["OutputKey"]: o["OutputValue"] for o in outputs}
            if "AutoScalingGroup" in outputs:
                info_json["InstanceDetails"] = get_instance_ip_from_scaling_group(
                    [outputs["AutoScalingGroup"]], ec2_deployment_config.region
                )
                info_json["Endpoints"] = get_endpoints_from_instance_address(
                    info_json["InstanceDetails"], api_names
                )
            if "S3Bucket" in outputs:
                info_json["S3Bucket"] = outputs["S3Bucket"]
            if "TargetGroup" in outputs:
                info_json["TargetGroup"] = outputs["TargetGroup"]
            if "Url" in outputs:
                info_json["Url"] = outputs["Url"]

            healthy_target = get_healthy_target(
                outputs["TargetGroup"], ec2_deployment_config.region
            )
            if healthy_target:
                deployment_state = DeploymentState.RUNNING
            else:
                deployment_state = DeploymentState.PENDING
            state = DeploymentState(
                state=deployment_state, info_json=json.dumps(info_json)
            )
            return DescribeDeploymentResponse(status=Status.OK(), state=state)

        except BentoMLException as error:
            return DescribeDeploymentResponse(status=error.status_proto)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# CSV utils following https://tools.ietf.org/html/rfc4180
from typing import Iterable, Iterator


def csv_splitlines(string) -> Iterator[str]:
    if '"' in string:

        def _iter_line(line):
            quoted = False
            last_cur = 0
            for i, c in enumerate(line):
                if c == '"':
                    quoted = not quoted
                if not quoted and string[i : i + 1] == "\n":
                    if i == 0 or string[i - 1] != '\r':
                        yield line[last_cur:i]
                        last_cur = i + 1
                    else:
                        yield line[last_cur : i - 1]
                        last_cur = i + 1
            yield line[last_cur:]

        return _iter_line(string)

    return iter(string.splitlines())


def csv_split(string, delimiter) -> Iterator[str]:
    if '"' in string:
        dlen = len(delimiter)

        def _iter_line(line):
            quoted = False
            last_cur = 0
            for i, c in enumerate(line):
                if c == '"':
                    quoted = not quoted
                if not quoted and string[i : i + dlen] == delimiter:
                    yield line[last_cur:i]
                    last_cur = i + dlen
            yield line[last_cur:]

        return _iter_line(string)
    else:
        return iter(string.split(delimiter))


def csv_row(tds: Iterable):
    return ",".join(csv_quote(td) for td in tds)


def csv_unquote(string):
    if '"' in string:
        string = string.strip()
        assert string[0] == '"' and string[-1] == '"'
        return string[1:-1].replace('""', '"')
    return string


def csv_quote(td):
    '''
    >>> csv_quote(1)
    '1'
    >>> csv_quote('string')
    'string'
    >>> csv_quote('a,b"c')
    '"a,b""c"'
    >>> csv_quote(' ')
    '" "'
    '''
    if td is None:
        td = ''
    elif not isinstance(td, str):
        td = str(td)
    if '\n' in td or '"' in td or ',' in td or not td.strip():
        return td.replace('"', '""').join('""')
    return td



import logging
import os
import pathlib
import shutil
import tempfile

from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv
from bentoml.utils.tensorflow import (
    cast_tensor_by_spec,
    get_arg_names,
    get_input_signatures,
    get_restored_functions,
    pretty_format_restored_model,
)

logger = logging.getLogger(__name__)


def _is_path_like(p):
    return isinstance(p, (str, bytes, pathlib.PurePath, os.PathLike))


class _TensorflowFunctionWrapper:
    """
    TensorflowFunctionWrapper
    transform input tensor following function input signature
    """

    def __init__(self, origin_func, arg_names=None, arg_specs=None, kwarg_specs=None):
        self.origin_func = origin_func
        self.arg_names = arg_names
        self.arg_specs = arg_specs
        self.kwarg_specs = {k: v for k, v in zip(arg_names or [], arg_specs or [])}
        self.kwarg_specs.update(kwarg_specs or {})

    def __call__(self, *args, **kwargs):
        if self.arg_specs is None and self.kwarg_specs is None:
            return self.origin_func(*args, **kwargs)

        for k in kwargs:
            if k not in self.kwarg_specs:
                raise TypeError(f"Function got an unexpected keyword argument {k}")

        arg_keys = {k for k, _ in zip(self.arg_names, args)}
        _ambiguous_keys = arg_keys & set(kwargs)
        if _ambiguous_keys:
            raise TypeError(f"got two values for arguments '{_ambiguous_keys}'")

        # INFO:
        # how signature with kwargs works?
        # https://github.com/tensorflow/tensorflow/blob/v2.0.0/tensorflow/python/eager/function.py#L1519

        transformed_args = tuple(
            cast_tensor_by_spec(arg, spec) for arg, spec in zip(args, self.arg_specs)
        )

        transformed_kwargs = {
            k: cast_tensor_by_spec(arg, self.kwarg_specs[k])
            for k, arg in kwargs.items()
        }
        return self.origin_func(*transformed_args, **transformed_kwargs)

    def __getattr__(self, k):
        return getattr(self.origin_func, k)

    @classmethod
    def hook_loaded_model(cls, loaded_model):
        funcs = get_restored_functions(loaded_model)
        for k, func in funcs.items():
            arg_names = get_arg_names(func)
            sigs = get_input_signatures(func)
            if not sigs:
                continue
            arg_specs, kwarg_specs = sigs[0]
            setattr(
                loaded_model,
                k,
                cls(
                    func,
                    arg_names=arg_names,
                    arg_specs=arg_specs,
                    kwarg_specs=kwarg_specs,
                ),
            )


def _load_tf_saved_model(path):
    try:
        import tensorflow as tf
        from tensorflow.python.training.tracking.tracking import AutoTrackable

        TF2 = tf.__version__.startswith('2')
    except ImportError:
        raise MissingDependencyException(
            "Tensorflow package is required to use TfSavedModelArtifact"
        )

    if TF2:
        return tf.saved_model.load(path)
    else:
        loaded = tf.compat.v2.saved_model.load(path)
        if isinstance(loaded, AutoTrackable) and not hasattr(loaded, "__call__"):
            logger.warning(
                '''Importing SavedModels from TensorFlow 1.x.
                `outputs = imported(inputs)` is not supported in bento service due to
                tensorflow API.

                Recommended usage:

                ```python
                from tensorflow.python.saved_model import signature_constants

                imported = tf.saved_model.load(path_to_v1_saved_model)
                wrapped_function = imported.signatures[
                    signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY]
                wrapped_function(tf.ones([]))
                ```

                See https://www.tensorflow.org/api_docs/python/tf/saved_model/load for
                details.
                '''
            )
        return loaded


class TensorflowSavedModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading Tensorflow model in tf.saved_model format

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: tensorflow package is required for
            TensorflowSavedModelArtifact

    Example usage:

    >>> import tensorflow as tf
    >>>
    >>> # Option 1: custom model with specific method call
    >>> class Adder(tf.Module):
    >>>     @tf.function(input_signature=[tf.TensorSpec(shape=None, dtype=tf.float32)])
    >>>     def add(self, x):
    >>>         return x + x + 1.
    >>> model_to_save = Adder()
    >>> # ... compiling, training, etc
    >>>
    >>> # Option 2: Sequential model (direct call only)
    >>> model_to_save = tf.keras.Sequential([
    >>>     tf.keras.layers.Flatten(input_shape=(28, 28)),
    >>>     tf.keras.layers.Dense(128, activation='relu'),
    >>>     tf.keras.layers.Dense(10, activation='softmax')
    >>> ])
    >>> # ... compiling, training, etc
    >>>
    >>> import bentoml
    >>> from bentoml.adapters import JsonInput
    >>> from bentoml.frameworks.tensorflow import TensorflowSavedModelArtifact
    >>>
    >>> @bentoml.env(pip_packages=["tensorflow"])
    >>> @bentoml.artifacts([TensorflowSavedModelArtifact('model')])
    >>> class TfModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=JsonInput(), batch=False)
    >>>     def predict(self, json):
    >>>         input_data = json['input']
    >>>         prediction = self.artifacts.model.add(input_data)
    >>>         # prediction = self.artifacts.model(input_data)  # if Sequential mode
    >>>         return prediction.numpy()
    >>>
    >>> svc = TfModelService()
    >>>
    >>> # Option 1: pack directly with Tensorflow trackable object
    >>> svc.pack('model', model_to_save)
    >>>
    >>> # Option 2: save to file path then pack
    >>> tf.saved_model.save(model_to_save, '/tmp/adder/1')
    >>> svc.pack('model', '/tmp/adder/1')
    """

    def __init__(self, name):
        super().__init__(name)

        self._model = None
        self._tmpdir = None
        self._path = None

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['tensorflow'])

    def _saved_model_path(self, base_path):
        return os.path.join(base_path, self.name + '_saved_model')

    def pack(
        self, obj, metadata=None, signatures=None, options=None
    ):  # pylint:disable=arguments-differ
        """
        Pack the TensorFlow Trackable object `obj` to [SavedModel format].
        Args:
          obj: A trackable object to export.
          signatures: Optional, either a `tf.function` with an input signature
            specified or the result of `f.get_concrete_function` on a
            `@tf.function`-decorated function `f`, in which case `f` will be used to
            generate a signature for the SavedModel under the default serving
            signature key. `signatures` may also be a dictionary, in which case it
            maps from signature keys to either `tf.function` instances with input
            signatures or concrete functions. The keys of such a dictionary may be
            arbitrary strings, but will typically be from the
            `tf.saved_model.signature_constants` module.
          options: Optional, `tf.saved_model.SaveOptions` object that specifies
            options for saving.

        Raises:
          ValueError: If `obj` is not trackable.
        """
        if not _is_path_like(obj):
            if self._tmpdir is not None:
                self._tmpdir.cleanup()
            else:
                self._tmpdir = tempfile.TemporaryDirectory()
            try:
                import tensorflow as tf

                TF2 = tf.__version__.startswith('2')
            except ImportError:
                raise MissingDependencyException(
                    "Tensorflow package is required to use TfSavedModelArtifact."
                )
            if TF2:
                tf.saved_model.save(
                    obj, self._tmpdir.name, signatures=signatures, options=options,
                )
            else:
                if self.options:
                    logger.warning(
                        "Parameter 'options: %s' is ignored when using Tensorflow "
                        "version 1",
                        str(options),
                    )
                tf.saved_model.save(
                    obj, self._tmpdir.name, signatures=signatures,
                )
            self._path = self._tmpdir.name
        else:
            self._path = obj
        self._packed = True
        loaded = self.get()
        logger.warning(
            "Due to TensorFlow's internal mechanism, only methods wrapped under "
            "`@tf.function` decorator and the Keras default function "
            "`__call__(inputs, training=False)` can be restored after a save & load.\n"
            "You can test the restored model object by referring:\n"
            f"<bento_svc>.artifacts.{self.name}\n"
        )
        logger.info(pretty_format_restored_model(loaded))
        if hasattr(loaded, "keras_api"):
            logger.warning(
                f"BentoML detected that {self.__class__.__name__} is being used "
                "to pack a Keras API based model. "
                "In order to get optimal serving performance, we recommend "
                f"either replacing {self.__class__.__name__} with KerasModelArtifact, "
                "or wrapping the keras_model.predict method with tf.function decorator."
            )
        return self

    def get(self):
        if self._model is None:
            loaded_model = _load_tf_saved_model(self._path)
            _TensorflowFunctionWrapper.hook_loaded_model(loaded_model)
            self._model = loaded_model
        return self._model

    def load(self, path):
        saved_model_path = self._saved_model_path(path)
        return self.pack(saved_model_path)

    def save(self, dst):
        # Copy exported SavedModel model directory to BentoML saved artifact directory
        shutil.copytree(self._path, self._saved_model_path(dst))

    def __del__(self):
        if getattr(self, "_tmpdir", None) is not None:
            self._tmpdir.cleanup()



import os
from importlib import import_module

from bentoml.service.env import BentoServiceEnv

from bentoml.exceptions import (
    InvalidArgument,
    MissingDependencyException,
    NotFound,
)
from bentoml.service import BentoServiceArtifact

try:
    import transformers
except ImportError:
    transformers = None


class TransformersModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading Transformers models

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: transformers package
            is required for TransformersModelArtifact
        InvalidArgument: invalid argument type, model being packed
            must be either a dictionary of format
            {'model':transformers model object,
            'tokenizer':transformers tokenizer object}
            or a directory path where the model is saved
            or a pre-trained model provided by transformers
            which can be loaded using transformers.AutoModelWithLMHead
        NotFound: if the provided model name or model path is not found

    Example usage:

    >>> import bentoml
    >>> from transformers import AutoModelWithLMHead, AutoTokenizer
    >>> from bentoml.adapters import JsonInput
    >>>
    >>> @bentoml.env(pip_packages=["transformers==3.1.0", "torch==1.6.0"])
    >>> @bentoml.artifacts([TransformersModelArtifact("gptModel")])
    >>> class TransformerService(bentoml.BentoService):
    >>>     @bentoml.api(input=JsonInput(), batch=False)
    >>>     def predict(self, parsed_json):
    >>>         src_text = parsed_json.get("text")
    >>>         model = self.artifacts.gptModel.get("model")
    >>>         tokenizer = self.artifacts.gptModel.get("tokenizer")
    >>>         input_ids = tokenizer.encode(src_text, return_tensors="pt")
    >>>         output = model.generate(input_ids, max_length=50)
    >>>         output = tokenizer.decode(output[0], skip_special_tokens=True)
    >>>         return output
    >>>
    >>>
    >>> ts = TransformerService()
    >>>
    >>> model_name = "gpt2"
    >>> model = AutoModelWithLMHead.from_pretrained("gpt2")
    >>> tokenizer = AutoTokenizer.from_pretrained("gpt2")
    >>> # Option 1: Pack using dictionary (recommended)
    >>> artifact = {"model": model, "tokenizer": tokenizer}
    >>> ts.pack("gptModel", artifact)
    >>> # Option 2: pack using the name of the model
    >>> # ts.pack("gptModel","gpt2")
    >>> # Note that while packing using the name of the model,
    >>> # ensure that the model can be loaded using
    >>> # transformers.AutoModelWithLMHead (eg GPT, Bert, Roberta etc.)
    >>> # If this is not the case (eg AutoModelForQuestionAnswering, BartModel etc)
    >>> # then pack the model by passing a dictionary
    >>> # with the model and tokenizer declared explicitly
    >>> saved_path = ts.save()
    """

    def __init__(self, name):
        super().__init__(name)
        self._model = None
        self._tokenizer_type = None
        self._model_type = "AutoModelWithLMHead"

        if transformers is None:
            raise MissingDependencyException(
                "transformers package is required to use TransformersModelArtifact"
            )

    def _file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def _load_from_directory(self, path):
        if self._model_type is None:
            raise NotFound(
                "Type of transformers model not found. "
                "This should be present in a file called "
                "'_model_type.txt' in the artifacts of the bundle."
            )
        if self._tokenizer_type is None:
            raise NotFound(
                "Type of transformers tokenizer not found. "
                "This should be present in a file called 'tokenizer_type.txt' "
                "in the artifacts of the bundle."
            )
        transformers_model = getattr(
            import_module("transformers"), self._model_type
        ).from_pretrained(path)
        tokenizer = getattr(
            import_module("transformers"), self._tokenizer_type
        ).from_pretrained(path)
        return {"model": transformers_model, "tokenizer": tokenizer}

    def _load_from_dict(self, model):
        if not model.get("model"):
            raise InvalidArgument(
                " 'model' key is not found in the dictionary."
                " Expecting a dictionary of with keys 'model' and 'tokenizer'"
            )
        if not model.get("tokenizer"):
            raise InvalidArgument(
                "'tokenizer' key is not found in the dictionary. "
                "Expecting a dictionary of with keys 'model' and 'tokenizer'"
            )

        model_class = str(type(model.get("model")).__module__)
        tokenizer_class = str(type(model.get("tokenizer")).__module__)
        # if either model or tokenizer is not an object of transformers
        if not model_class.startswith("transformers"):
            raise InvalidArgument(
                "Expecting a transformers model object but object passed is {}".format(
                    type(model.get("model"))
                )
            )
        if not tokenizer_class.startswith("transformers"):
            raise InvalidArgument(
                "Expecting a transformers model object but object passed is {}".format(
                    type(model.get("tokenizer"))
                )
            )
        # success
        return model

    def _load_from_string(self, model_name):
        try:
            transformers_model = getattr(
                import_module("transformers"), self._model_type
            ).from_pretrained(model_name)
            tokenizer = transformers.AutoTokenizer.from_pretrained(model_name)
            return {"model": transformers_model, "tokenizer": tokenizer}
        except EnvironmentError:
            raise NotFound(
                "model with the name {} is not present "
                "in the transformers library".format(model_name)
            )
        except AttributeError:
            raise NotFound(
                "transformers has no model type called {}".format(self._model_type)
            )

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['xgboost'])

    def pack(self, model, metadata=None):
        loaded_model = None
        if isinstance(model, str):
            if os.path.isdir(model):
                loaded_model = self._load_from_directory(model)
            else:
                loaded_model = self._load_from_string(model)
        elif isinstance(model, dict):
            loaded_model = self._load_from_dict(model)
        else:
            raise InvalidArgument(
                "Expecting a Dictionary of format "
                "{'model':<transformers model object>,'tokenizer':<tokenizer object> }"
            )
        self._model = loaded_model
        return self

    def load(self, path):
        path = self._file_path(path)
        with open(os.path.join(path, "_model_type.txt"), "r") as f:
            self._model_type = f.read().strip()
        with open(os.path.join(path, "tokenizer_type.txt"), "r") as f:
            self._tokenizer_type = f.read().strip()
        return self.pack(path)

    def _save_model_type(self, path):
        with open(os.path.join(path, "_model_type.txt"), "w") as f:
            f.write(self._model_type)
        with open(os.path.join(path, "tokenizer_type.txt"), "w") as f:
            f.write(self._tokenizer_type)

    def save(self, dst):
        path = self._file_path(dst)
        self._model_type = self._model.get("model").__class__.__name__
        self._tokenizer_type = self._model.get("tokenizer").__class__.__name__
        self._model.get("model").save_pretrained(path)
        self._model.get("tokenizer").save_pretrained(path)
        self._save_model_type(path)
        return path

    def get(self):
        return self._model



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging.config
import os
from pathlib import Path
import sys

from simple_di import Provide, inject

from bentoml.configuration import get_debug_mode
from bentoml.configuration.containers import BentoMLContainer


def get_logging_config_dict(
    logging_level: str,
    base_log_directory: str,
    console_logging_enabled: bool,
    file_logging_enabled: bool,
):
    MEGABYTES = 1024 * 1024

    handlers = {}
    bentoml_logger_handlers = []
    prediction_logger_handlers = []
    feedback_logger_handlers = []
    if console_logging_enabled:
        handlers.update(
            {
                "console": {
                    "level": logging_level,
                    "formatter": "console",
                    "class": "logging.StreamHandler",
                    "stream": sys.stdout,
                }
            }
        )
        bentoml_logger_handlers.append("console")
        prediction_logger_handlers.append("console")
        feedback_logger_handlers.append("console")
    if file_logging_enabled:
        handlers.update(
            {
                "local": {
                    "level": logging_level,
                    "formatter": "dev",
                    "class": "logging.handlers.RotatingFileHandler",
                    "filename": os.path.join(base_log_directory, "active.log"),
                    "maxBytes": 100 * MEGABYTES,
                    "backupCount": 2,
                },
                "prediction": {
                    "class": "logging.handlers.RotatingFileHandler",
                    "formatter": "prediction",
                    "level": "INFO",
                    "filename": os.path.join(base_log_directory, "prediction.log"),
                    "maxBytes": 100 * MEGABYTES,
                    "backupCount": 10,
                },
                "feedback": {
                    "class": "logging.handlers.RotatingFileHandler",
                    "formatter": "feedback",
                    "level": "INFO",
                    "filename": os.path.join(base_log_directory, "feedback.log"),
                    "maxBytes": 100 * MEGABYTES,
                    "backupCount": 10,
                },
            }
        )
        bentoml_logger_handlers.append("local")
        prediction_logger_handlers.append("prediction")
        feedback_logger_handlers.append("feedback")

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "console": {"format": "[%(asctime)s] %(levelname)s - %(message)s"},
            "dev": {
                "format": "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s"
                " - %(message)s"
            },
            "prediction": {"()": "pythonjsonlogger.jsonlogger.JsonFormatter"},
            "feedback": {"()": "pythonjsonlogger.jsonlogger.JsonFormatter"},
        },
        "handlers": handlers,
        "loggers": {
            "bentoml": {
                "handlers": bentoml_logger_handlers,
                "level": logging_level,
                "propagate": False,
            },
            "bentoml.prediction": {
                "handlers": prediction_logger_handlers,
                "level": "INFO",
                "propagate": False,
            },
            "bentoml.feedback": {
                "handlers": feedback_logger_handlers,
                "level": "INFO",
                "propagate": False,
            },
        },
    }


@inject
def configure_logging(
    logging_level: str = Provide[BentoMLContainer.config.logging.level],
    base_log_dir: str = Provide[BentoMLContainer.logging_file_directory],
    console_logging_enabled: bool = Provide[
        BentoMLContainer.config.logging.console.enabled
    ],
    file_logging_enabled: bool = Provide[BentoMLContainer.config.logging.file.enabled],
    advanced_enabled: bool = Provide[BentoMLContainer.config.logging.advanced.enabled],
    advanced_config: dict = Provide[BentoMLContainer.config.logging.advanced.config],
):
    Path(base_log_dir).mkdir(parents=True, exist_ok=True)
    if advanced_enabled:
        logging.config.dictConfig(advanced_config)
        logging.getLogger(__name__).debug(
            "Configured logging with advanced configuration, config=%s", advanced_config
        )
    else:
        if get_debug_mode():
            logging_level = logging.getLevelName(logging.DEBUG)

        logging_config = get_logging_config_dict(
            logging_level, base_log_dir, console_logging_enabled, file_logging_enabled
        )
        logging.config.dictConfig(logging_config)
        logging.getLogger(__name__).debug(
            "Configured logging with simple configuration, "
            "level=%s, directory=%s, console_enabled=%s, file_enabled=%s",
            logging_level,
            base_log_dir,
            console_logging_enabled,
            file_logging_enabled,
        )
